from medspacy.visualization import visualize_ent

DEFAULT_PIPENAMES = {
    "sentencizer",
    "target_matcher",
    "context",
    "tokenizer",
}

ALL_PIPE_NAMES = {
    "sentencizer",
    "target_matcher",
    "context",
    "tokenizer",
    "preprocessor",
    "postprocessor",
    "sectionizer"
}


def load(model="default", enable=None, disable=None, load_rules=True):
    """Load a spaCy language object with medSpaCy pipeline components.
    By default, the base model will be 'en_core_web_sm' with the
    following components:
        - "tokenizer": A customized tokenizer (set to be nlp.tokenizer)
        - "sentencizer": PyRuSH Sentencizer for sentence splitting
        - "target_matcher": TargetMatcher for extended pattern matching
        - "context": ConText for attribute assertion
    Args:
        model: The name of the base spaCy model to load. Default 'language' will load the tagger and parser
            from "en_core_web_sm".
        enable (iterable, str, or None): A string or list of component names to include in the pipeline.
            If None, will include all pipeline components listed above.
            If "all", will load all medspaCy components.
            If a list or other iterable, will load the specified pipeline components.
            Note that if using enable, *all* desired pipeline component names must be included.
            Pipeline components could also be instantiated separately and added using the `nlp.add_pipe` method.
            In addition to the default values above, the following medspaCy components may also be added
            and will be added if enable="all":
                - "sectionizer": A SectionDetection component.
                    See medspacy.section_detection.Sectionizer
                - "preprocessor": A wrapper around the tokenizer for destructive preprocessing.
                    Rules added to the preprocessor will modify the underlying text.
                    This component will be set as nlp.tokenizer and will not be listed with nlp.pipe_names.
                    See medspacy.preprocess.Preprocessor
                - "postprocessor": A component for implementing custom business logic at the end of the pipeline
                    and modifying entities by removing them from doc.ents or setting attributes.
                    See medspacy.postprocess.Postprocessor
            Any additional component names (ie., not a medspaCy component) will be passed into
                spacy.load(model_name, enable=enable) and will apply to the base model.
                ie., medspacy.load("en_core_web_sm", enable=["tagger", "parser", "context"])
                will load the tagger and parser from en_core_web_sm and then add context.
        disable (iterable or None): A list of component names to exclude.
            Cannot be set if `enable` is not None.
        load_rules (bool): Whether or not to include default rules for available components.
            If True, sectionizer and context will both be loaded with default rules.
            Default is True.

    Returns:
        nlp: a spaCy Language object
    """

    model, enable, disable = _build_pipe_names(model, enable, disable)
    import spacy
    nlp = spacy.load(model, disable=disable)

    if "tokenizer" in enable:
        from .custom_tokenizer import create_medspacy_tokenizer
        medspacy_tokenizer = create_medspacy_tokenizer(nlp)
        nlp.tokenizer = medspacy_tokenizer

    if "preprocessor" in enable:
        from .preprocess import Preprocessor
        preprocessor = Preprocessor(nlp.tokenizer)
        nlp.tokenizer = preprocessor


    if "sentencizer" in enable:
        from os import path
        from pathlib import Path

        pyrush_path = path.join(
            Path(__file__).resolve().parents[1], "resources", "rush_rules.tsv"
        )
        from .sentence_splitting import PyRuSHSentencizer

        pyrush = PyRuSHSentencizer(pyrush_path)
        if "parser" in nlp.pipe_names:
            if "tagger" in nlp.pipe_names:
                nlp.add_pipe(pyrush, before="tagger")
            else:
                nlp.add_pipe(pyrush, before="parser")
        else:
            nlp.add_pipe(pyrush)

    if "target_matcher" in enable:
        from .ner import TargetMatcher

        target_matcher = TargetMatcher(nlp)
        nlp.add_pipe(target_matcher)

    if "context" in enable:
        from .context import ConTextComponent

        if load_rules:
            context = ConTextComponent(nlp, rules="default")
        else:
            context = ConTextComponent(nlp, rules=None)
        nlp.add_pipe(context)

    if "sectionizer" in enable:
        from .section_detection import Sectionizer

        if load_rules:
            sectionizer = Sectionizer(nlp, patterns="default")
        else:
            sectionizer = Sectionizer(nlp, patterns=None)
        nlp.add_pipe(sectionizer)

    if "postprocessor" in enable:
        from .postprocess import Postprocessor
        postprocessor = Postprocessor()
        nlp.add_pipe(postprocessor)

    return nlp


def _build_pipe_names(model, enable=None, disable=None):
    """Implement logic based on the pipenames defined in 'enable' and 'disable'.
    If enable and disable are both None, then it will load the default pipenames.
    Otherwise, will allow custom selection of components.
    """
    if enable is not None and disable is not None:
        raise ValueError("Either `enable` or `disable` must be None.")

    if disable is not None:
        # If there's a single pipe name, next it in a set
        if isinstance(disable, str):
            disable = {disable}
        else:
            disable = set(disable)
        enable = DEFAULT_PIPENAMES.difference(set(disable))
    elif enable is not None:

        if isinstance(enable, str):
            if enable == "all":
                enable = ALL_PIPE_NAMES.copy()
            else:
                enable = {enable}
        else:
            enable = set(enable)
        disable = set(DEFAULT_PIPENAMES).difference(enable)
    else:
        enable = DEFAULT_PIPENAMES
        disable = set()
    # We'll eventually have an actual medSpaCy model here
    # but for now we're basing it off of "en_core_web_sm"
    if model == "default":
        model = "en_core_web_sm"
        disable.update({"ner", "tagger", "parser"})
    return model, enable, disable



import json


class ConTextItem:
    """An ConTextItem defines a ConText modifier. ConTextItems are rules define
    which spans are extracted as modifiers and how they behave, such as the phrase to be matched,
    the category/semantic class, the direction of the modifier in the text, and what types of target
    spans can be modfified.
    """

    _ALLOWED_RULES = (
        "FORWARD",
        "BACKWARD",
        "BIDIRECTIONAL",
        "TERMINATE",
        "PSEUDO",
    )
    _ALLOWED_KEYS = {
        "literal",
        "rule",
        "pattern",
        "category",
        "metadata",
        "allowed_types",
        "excluded_types",
        "max_targets",
        "max_scope",
    }

    def __init__(
        self,
        literal,
        category,
        rule="BIDIRECTIONAL",
        pattern=None,
        on_match=None,
        on_modifies=None,
        allowed_types=None,
        excluded_types=None,
        max_scope=None,
        max_targets=None,
        terminated_by=None,
        metadata=None,
        filtered_types=None,
    ):
        """Create an ConTextItem object.
        The primary arguments of `literal` `category`, and `rule` define
        the span of text to be matched, the semantic category, and the direction
        within the sentence in which the modifier operates.
        Other arguments specify additional custom logic such as:
            - Additional control over what text can be matched as a modifier (pattern and on_match)
            - Which types of targets can be modified (allowed_types, excluded_types)
            - The scope size and number of targets that a modifier can modify (max_targets, max_scope)
            - Other logic for terminating a span or for allowing a modifier to modify a target (on_modifies, terminated_by)


        Args:
            literal (str): The actual string of a concept. If pattern is None,
                this string will be lower-cased and matched to the lower-case string.
            category (str): The semantic class of the item.
            pattern (list or None): A spaCy pattern to match using token attributes.
                See https://spacy.io/usage/rule-based-matching.
            rule (str): The directionality or action of a modifier. This defines which part
                of a sentence a modifier will include as its scope. Entities within
                the scope will be considered to be modified.
                Valid values are:
                 - "FORWARD": Scope will begin after the end of a modifier and move
                    to the right towards the end of the sentence
                - "BACKWARD": Scope will begin before the beginning of a modifier
                    and move to the left towards the beginning of a sentence
                - "BIDIRECTIONAL": Scope will expand on either side of a modifier
                - "TERMINATE": A special rule to limit any other modifiers
                    if this phrase is in its scope.
                    Example: "no evidence of chf but there is pneumonia":
                        "but" will prevent "no evidence of" from modifying "pneumonia"
                - "PSEUDO": A special rule which will not modify any targets.
                    This can be used for differentiating superstrings of modifiers.
                    Example: A modifier with literal="negative attitude"
                    will prevent the phrase "negative" in "She has a negative attitude about her treatment"
                    from being extracted as a modifier.
            on_match (callable or None): Callback function to act on spaCy matches.
                Takes the argument matcher, doc, i, and matches.
            on_modifies (callable or None): Callback function to run when building an edge
                between a target and a modifier. This allows specifying custom logic for
                allowing or preventing certain modifiers from modifying certain targets.
                The callable should take 3 arguments:
                    target: The spaCy Span from doc.ents (ie., 'Evidence of pneumonia')
                    modifier: The spaCy Span covered in a resulting modifier (ie., 'no evidence of')
                    span_between: The Span between the target and modifier in question.
                Should return either True or False. If returns False, then the modifier will not modify
                the target.
#           allowed_types (set or None): A set of target labels to allow a modifier to modify.
                If None, will apply to any type not specifically excluded in excluded_types.
                Only one of allowed_types and excluded_types can be used. An error will be thrown
                if both are not None.
            excluded_types (set or None): A set of target labels which this modifier cannot modify.
                If None, will apply to all target types unless allowed_types is not None.
            max_scope (int or None): A number of tokens to explicitly limit the size of the modifier's scope.
                If None, the scope will include the entire sentence in the direction of `rule`
                (ie., starting at the beginning of the sentence for "BACKWARD", end of the sentence for "FORWARD",
                and the entire sentence for "BIDIRECTIONAL".
                This is useful for requiring modifiers be very close to a concept in the text
                or for preventing too far of modifier ranges caused by undersplitting of sentences.
                Example: In the sentence "pt with chf, pna vs. rsv", "vs." will modify
                    only "pna" and "rsv" if max_scope=1 and rule="BIDIRECTIONAL".
            max_targets (int or None): The maximum number of targets which a modifier can modify.
                If None, will modify all targets in its scope.
            terminated_by (iterable or None): An optional array of other modifier categories which will
                terminate the scope of this modifier. If None, only "TERMINATE" will do this.
                Example: if a ConTextItem defining "positive for" has terminated_by={"NEGATED_EXISTENCE"},
                then in the sentence "positive for flu, negative for RSV", the positive modifier
                will modify "flu" but will be terminated by "negative for" and will not modify "RSV".
                This helps prevent multiple conflicting modifiers from distributing too far across
                a sentence.
            metadata (dict or None): A dict of additional data to pass in,
                such as free-text comments, additional attributes, or ICD-10 codes.
                Default None.

        Returns:
            item: a ConTextItem
        """
        self.literal = literal.lower()
        self.category = category.upper()
        self.pattern = pattern
        self.rule = rule.upper()
        self.on_match = on_match
        self.on_modifies = on_modifies

        if allowed_types is not None and excluded_types is not None:
            raise ValueError(
                "A ConTextItem was instantiated with non-null values for both allowed_types and excluded_types. "
                "Only one of these can be non-null, since cycontext either explicitly includes or excludes target types."
            )
        if allowed_types is not None:
            self.allowed_types = {label.upper() for label in allowed_types}
        else:
            self.allowed_types = None
        if excluded_types is not None:
            self.excluded_types = {label.upper() for label in excluded_types}
        else:
            self.excluded_types = None

        if max_targets is not None and max_targets <= 0:
            raise ValueError("max_targets must be >= 0 or None.")
        self.max_targets = max_targets
        if max_scope is not None and max_scope <= 0:
            raise ValueError("max_scope must be >= 0 or None.")
        self.max_scope = max_scope
        if terminated_by is None:
            terminated_by = set()
        else:
            if isinstance(terminated_by, str):
                raise ValueError(
                    "terminated_by must be an iterable, such as a list or set, not {}.".format(
                        terminated_by
                    )
                )
            terminated_by = {string.upper() for string in terminated_by}

        self.terminated_by = terminated_by

        self.filtered_types = filtered_types

        self.metadata = metadata

        if self.rule not in self._ALLOWED_RULES:
            raise ValueError(
                "Rule {0} not recognized. Must be one of: {1}".format(
                    self.rule, self._ALLOWED_RULES
                )
            )

    @classmethod
    def from_yaml(cls, _file):
        """Read in a lexicon of modifiers from a YAML file.

        Args:
            filepath: the .yaml file containing modifier rules

        Returns:
            context_item: a list of ConTextItem objects
        Raises:
            KeyError: if the dictionary contains any keys other than
                those accepted by ConTextItem.__init__
        """

        import yaml
        import urllib.request, urllib.error, urllib.parse

        def _get_fileobj(_file):
            if not urllib.parse.urlparse(_file).scheme:
                _file = "file://" + _file
            return urllib.request.urlopen(_file, data=None)

        f0 = _get_fileobj(_file)
        context_items = [
            ConTextItem.from_dict(data) for data in yaml.safe_load_all(f0)
        ]
        f0.close()
        return {"item_data": context_items}

    @classmethod
    def from_json(cls, filepath):
        """Read in a lexicon of modifiers from a JSON file.

        Args:
            filepath: the .json file containing modifier rules

        Returns:
            context_item: a list of ConTextItem objects
        Raises:
            KeyError: if the dictionary contains any keys other than
                those accepted by ConTextItem.__init__
        """

        with open(filepath) as file:
            modifier_data = json.load(file)
        item_data = []
        for data in modifier_data["item_data"]:
            item_data.append(ConTextItem.from_dict(data))
        return item_data

    @classmethod
    def from_dict(cls, item_dict):
        """Reads a dictionary into a ConTextItem. Used when reading from a json file.

        Args:
            item_dict: the dictionary to convert

        Returns:
            item: the ConTextItem created from the dictionary

        Raises:
            ValueError: if the json is invalid
        """
        try:
            item = ConTextItem(**item_dict)
        except Exception as err:
            print(err)
            keys = set(item_dict.keys())
            invalid_keys = keys.difference(cls._ALLOWED_KEYS)
            msg = (
                "JSON object contains invalid keys: {0}.\n"
                "Must be one of: {1}".format(invalid_keys, cls._ALLOWED_KEYS)
            )
            raise ValueError(msg)

        return item

    @classmethod
    def to_json(cls, item_data, filepath):
        """Writes ConTextItems to a json file.

        Args:
            item_data: a list of ConTextItems that will be written to a file.
            filepath: the .json file to contain modifier rules
        """

        data = {"item_data": [item.to_dict() for item in item_data]}
        with open(filepath, "w") as file:
            json.dump(data, file, indent=4)

    def to_dict(self):
        """Converts ConTextItems to a python dictionary. Used when writing context items to a json file.

        Returns:
            item_dict: the dictionary containing the ConTextItem info.
        """
        item_dict = {}
        for key in self._ALLOWED_KEYS:
            item_dict[key] = self.__dict__.get(key)
        return item_dict

    def __repr__(self):
        return f"ConTextItem(literal='{self.literal}', category='{self.category}', pattern={self.pattern}, rule='{self.rule}')"



class PostprocessingRule:

    def __init__(self, patterns, action, name=None, description=None, action_args=None):
        """A PostprocessingRule checks conditions of a spaCy Span entity
        and executes some action if all patterns are met.

        patterns (list): A list of PostprocessingPatterns,
            each of which check a condition of an entity.
        action (function): A function to call with the entity as an argument.
            This function should take ay least the following two arguments:
                ent: the spacy span
                i: the index of ent in doc.ents
            Additional positional arguments can be provided in action_args.
        name (str): Optional name of rule.
        description (str): Optional description of the rule.
        action_args (tuple or None): Optional tuple of positional arguments
            to pass to action() if all patterns pass. Default is None,
            in which case the rule will call action(ent, i).

        """
        self.patterns = patterns
        self.action = action
        self.name = name
        self.description = description
        self.action_args = action_args

    def __call__(self, ent, i, debug=False):
        """Iterate through all of the patterns in self.rules.
        If any pattern does not pass (ie., return True), then returns False.
        If they all pass, execute self.action and return True.
        """
        for pattern in self.patterns:
            # If this is a tuple, at least one has to pass
            if isinstance(pattern, tuple):
                passed = False
                for subpattern in pattern:
                    rslt = subpattern(ent)
                    if rslt is True:
                        passed = True
                        break
                if passed is False:
                    return False
            # Otherwise just check a single value
            else:
                rslt = pattern(ent)
                if rslt is False:
                    return False

        # Every pattern passed - do the action
        if debug:
            print("Passed:", self, "on ent:", ent, ent.sent)
        if self.action_args is None:
            try:
                self.action(ent, i)
            except TypeError:
                _raise_action_error(self.action, (ent, i))
        else:
            try:
                self.action(ent, i, *self.action_args)
            except TypeError:
                _raise_action_error(self.action, (ent, i, *self.action_args))
        return True

    def __repr__(self):
        return f"PostprocessingRule: {self.name} - {self.description}"

def _raise_action_error(func, args):
    raise ValueError("The action function {0} does not have the correct number of arguments. "
                     "Any action function must start with two arguments: (ent, i) - the span and the index of "
                     "the span in doc.ents. Any additional arguments must be provided in a tuple "
                     "in `rule.action_args`. "
                     "Actual arguments passed in: {1} ".format(func, args))


from spacy.tokens import Doc, Token, Span
from spacy.matcher import Matcher, PhraseMatcher

# Filepath to default rules which are included in package
from os import path
from pathlib import Path
import re
import warnings

from . import util

Doc.set_extension("sections", default=list(), force=True)
Doc.set_extension("section_titles", getter=util.get_section_titles, force=True)
Doc.set_extension(
    "section_headers", getter=util.get_section_headers, force=True
)
Doc.set_extension("section_spans", getter=util.get_section_spans, force=True)
Doc.set_extension(
    "section_parents", getter=util.get_section_parents, force=True
)

Token.set_extension("section_span", default=None, force=True)
Token.set_extension("section_title", default=None, force=True)
Token.set_extension("section_header", default=None, force=True)
Token.set_extension("section_parent", default=None, force=True)

# Set span attributes to the attribute of the first token
# in case there is some overlap between a span and a new section header
Span.set_extension(
    "section_span", getter=lambda x: x[0]._.section_span, force=True
)
Span.set_extension(
    "section_title", getter=lambda x: x[0]._.section_title, force=True
)
Span.set_extension(
    "section_header", getter=lambda x: x[0]._.section_header, force=True
)
Span.set_extension(
    "section_parent", getter=lambda x: x[0]._.section_header, force=True
)

DEFAULT_RULES_FILEPATH = path.join(
    Path(__file__).resolve().parents[2],
    "resources",
    "section_patterns.jsonl",
)

DEFAULT_ATTRS = {
    "past_medical_history": {"is_historical": True},
    "sexual_and_social_history": {"is_historical": True},
    "family_history": {"is_family": True},
    "patient_instructions": {"is_hypothetical": True},
    "education": {"is_hypothetical": True},
    "allergy": {"is_hypothetical": True},
}
from collections import namedtuple
Section = namedtuple("Section", field_names=["section_title",
                                            "section_header",
                                            "section_parent",
                                            "section_span"])


class Sectionizer:
    name = "sectionizer"

    def __init__(
        self,
        nlp,
        patterns="default",
        add_attrs=False,
        max_scope=None,
        phrase_matcher_attr="LOWER",
        require_start_line=False,
        require_end_line=False,
        newline_pattern=r"[\n\r]+[\s]*$",
    ):
        """Create a new Sectionizer component. The sectionizer will search for spans in the text which
        match section header patterns, such as 'Past Medical History:'. Sections will be represented
        in custom attributes as:
            section_title (str): A normalized title of the section. Example: 'past_medical_history'
            section_header (Span): The Span of the doc which was matched as a section header.
                Example: 'Past Medical History:'
            section (Span): The entire section of the note, starting with section_header and up until the end
                of the section, which will be either the start of the next section header of some pre-specified
                scope. Example: 'Past Medical History: Type II DM'

        Section attributes will be registered for each Doc, Span, and Token in the following attributes:
            Doc._.sections: A list of namedtuples of type Section with 4 elements:
                - section_title
                - section_header
                - section_parent
                - section_span.
            A Doc will also have attributes corresponding to lists of each
                (ie., Doc._.section_titles, Doc._.section_headers, Doc._.section_parents, Doc._.section_spans)
            (Span|Token)._.section_title
            (Span|Token)._.section_header
            (Span|Token)._.section_parent
            (Span|Token)._.section_span

        Args:
            nlp: A SpaCy language model object
            patterns (str, list, or None): Where to read patterns from. Default is "default", which will
                load the default patterns provided by medSpaCy, which are derived from MIMIC-II.
                If a list, should be a list of pattern dicts following these conventional spaCy formats:
                    [
                        {"section_title": "past_medical_history", "pattern": "Past Medical History:"},
                        {"section_title": "problem_list", "pattern": [{"TEXT": "PROBLEM"}, {"TEXT": "LIST"}, {"TEXT": ":"}]}
                    ]
                If a string other than "default", should be a path to a jsonl file containing patterns.
            max_scope (None or int): Optional argument specifying the maximum number of tokens following a section header
                which can be included in a section. This can be useful if you think your section patterns are incomplete
                and want to prevent sections from running too long in the note. Default is None, meaning that the scope
                of a section will be until either the next section header or the end of the document.
            phrase_matcher_attr (str): The name of the token attribute which will be used by the PhraseMatcher
                for any patterns with a "pattern" value of a string.
            require_start_line (bool): Optionally require a section header to start on a new line. Default False.
            require_end_line (bool): Optionally require a section header to end with a new line. Default False.
            newline_pattern (str): Regular expression to match the new line either preceding or following a header
                if either require_start_line or require_end_line are True.
        """
        self.nlp = nlp
        self.add_attrs = add_attrs
        self.matcher = Matcher(nlp.vocab)
        self.max_scope = max_scope
        self.phrase_matcher = PhraseMatcher(
            nlp.vocab, attr=phrase_matcher_attr
        )
        self.require_start_line = require_start_line
        self.require_end_line = require_end_line
        self.newline_pattern = re.compile(newline_pattern)
        self.assertion_attributes_mapping = None
        self._patterns = []
        self._section_titles = set()
        self._parent_sections = {}
        self._parent_required = {}

        if patterns is not None:
            if patterns == "default":
                import os

                if not os.path.exists(DEFAULT_RULES_FILEPATH):
                    raise FileNotFoundError(
                        "The expected location of the default patterns file cannot be found. Please either "
                        "add patterns manually or add a jsonl file to the following location: ",
                        DEFAULT_RULES_FILEPATH,
                    )
                self.add(self.load_patterns_from_jsonl(DEFAULT_RULES_FILEPATH))
            # If a list, add each of the patterns in the list
            elif isinstance(patterns, list):
                self.add(patterns)
            elif isinstance(patterns, str):
                import os

                assert os.path.exists(patterns)
                self.add(self.load_patterns_from_jsonl(patterns))

        if add_attrs is False:
            self.add_attrs = False
        elif add_attrs is True:
            self.assertion_attributes_mapping = DEFAULT_ATTRS
            self.register_default_attributes()
        elif isinstance(add_attrs, dict):
            # Check that each of the attributes being added has been set
            for modifier in add_attrs.keys():
                attr_dict = add_attrs[modifier]
                for attr_name, attr_value in attr_dict.items():
                    if not Span.has_extension(attr_name):
                        raise ValueError(
                            "Custom extension {0} has not been set. Call Span.set_extension."
                        )

            self.add_attrs = True
            self.assertion_attributes_mapping = add_attrs

        else:
            raise ValueError(
                "add_attrs must be either True (default), False, or a dictionary, not {0}".format(
                    add_attrs
                )
            )

    @property
    def patterns(self):
        return self._patterns

    @property
    def section_titles(self):
        return self._section_titles

    @classmethod
    def load_patterns_from_jsonl(self, filepath):

        import json

        patterns = []
        with open(filepath) as f:
            for line in f:
                if line.startswith("//"):
                    continue
                patterns.append(json.loads(line))

        return patterns

    def register_default_attributes(self):
        """Register the default values for the Span attributes defined in DEFAULT_ATTRS."""
        for attr_name in [
            "is_negated",
            "is_uncertain",
            "is_historical",
            "is_hypothetical",
            "is_family",
        ]:
            try:
                Span.set_extension(attr_name, default=False)
            except ValueError:  # Extension already set
                pass

    def add(self, patterns):
        """Add a list of patterns to the clinical_sectionizer. Each pattern should be a dictionary with
       two keys:
           'section': The normalized section name of the section, such as 'pmh'.
           'pattern': The spaCy pattern matching a span of text.
               Either a string for exact matching (case insensitive)
               or a list of dicts.

       Example:
       >>> patterns = [ \
           {"section_title": "past_medical_history", "pattern": "pmh"}\
           {"section_title": "past_medical_history", "pattern": [{"LOWER": "past", "OP": "?"}, \
               {"LOWER": "medical"}, \
               {"LOWER": "history"}]\
               },\
           {"section_title": "assessment_and_plan", "pattern": "a/p:"}\
           ]
       >>> clinical_sectionizer.add(patterns)
       """
        for pattern_dict in patterns:
            name = pattern_dict["section_title"]
            pattern = pattern_dict["pattern"]
            parents = []
            parent_required = False
            if "parents" in pattern_dict.keys():
                parents = pattern_dict["parents"]

            if "parent_required" in pattern_dict.keys():
                if not parents:
                    raise ValueError(
                        "Jsonl file incorrectly formatted for pattern name {0}. If parents are required, then at least one parent must be specified.".format(
                            name
                        )
                    )
                parent_required = pattern_dict["parent_required"]

            if isinstance(pattern, str):
                self.phrase_matcher.add(name, None, self.nlp.make_doc(pattern))
            else:
                self.matcher.add(name, [pattern])
            self._patterns.append(pattern_dict)
            self._section_titles.add(name)

            if name in self._parent_sections.keys() and parents != []:
                warnings.warn(
                    "Duplicate section title {0}. Merging parents. If this is not indended, please specify distinc titles.".format(
                        name
                    ),
                    RuntimeWarning,
                )
                self._parent_sections[name].update(parents)
            else:
                self._parent_sections[name] = set(parents)

            if (
                name in self._parent_required.keys()
                and self._parent_required[name] != parent_required
            ):
                warnings.warn(
                    "Duplicate section title {0} has different parent_required option. Setting parent_required to False.".format(
                        name
                    ),
                    RuntimeWarning,
                )
                self._parent_required[name] = False
            else:
                self._parent_required[name] = parent_required

    def set_parent_sections(self, sections):
        """Determine the legal parent-child section relationships from the list
        of in-order sections of a document and the possible parents of each
        section as specified during rule creation.

        Args:
            sections: a list of spacy match tuples found in the doc
        """
        sections_final = []
        removed_sections = 0
        for i, (match_id, start, end) in enumerate(sections):
            name = self.nlp.vocab.strings[match_id]
            required = self._parent_required[name]
            i_a = i - removed_sections  # adjusted index for removed values
            if required and i_a == 0:
                removed_sections += 1
                continue
            elif i_a == 0 or name not in self._parent_sections.keys():
                sections_final.append((match_id, start, end, None))
            else:
                parents = self._parent_sections[name]
                identified_parent = None
                for parent in parents:
                    # go backwards through the section "tree" until you hit a root or the start of the list
                    candidate = self.nlp.vocab.strings[
                        sections_final[i_a - 1][0]
                    ]
                    candidates_parent = sections_final[i_a - 1][3]
                    candidate_i = i_a - 1
                    while candidate:
                        if candidate == parent:
                            identified_parent = parent
                            candidate = None
                        else:
                            # if you are at the end of the list... no parent
                            if candidate_i < 1:
                                candidate = None
                                continue
                            # if the current candidate has no parent... no parent exists
                            if not candidates_parent:
                                candidate = None
                                continue
                            # otherwise get the previous item in the list
                            temp = self.nlp.vocab.strings[
                                sections_final[candidate_i - 1][0]
                            ]
                            temp_parent = sections_final[candidate_i - 1][3]
                            # if the previous item is the parent of the current item
                            # OR if the previous item is a sibling of the current item
                            # continue to search
                            if (
                                temp == candidates_parent
                                or temp_parent == candidates_parent
                            ):
                                candidate = temp
                                candidates_parent = temp_parent
                                candidate_i -= 1
                            # otherwise, there is no further tree traversal
                            else:
                                candidate = None

                # if a parent is required, then add
                if identified_parent or not required:
                    # if the parent is identified, add section
                    # if the parent is not required, add section
                    # if parent is not identified and required, do not add the section
                    sections_final.append(
                        (match_id, start, end, identified_parent)
                    )
                else:
                    removed_sections += 1
        return sections_final

    def set_assertion_attributes(self, ents):
        """Add Span-level attributes to entities based on which section they occur in.

        Args:
            edges: the edges to modify

        """
        for ent in ents:
            if ent._.section_title in self.assertion_attributes_mapping:
                attr_dict = self.assertion_attributes_mapping[
                    ent._.section_title
                ]
                for (attr_name, attr_value) in attr_dict.items():
                    setattr(ent._, attr_name, attr_value)

    def __call__(self, doc):
        matches = self.matcher(doc)
        matches += self.phrase_matcher(doc)
        if self.require_start_line:
            matches = self.filter_start_lines(doc, matches)
        if self.require_end_line:
            matches = self.filter_end_lines(doc, matches)
        matches = prune_overlapping_matches(matches)
        matches = self.set_parent_sections(matches)
        # If this has already been processed by the sectionizer, reset the sections
        doc._.sections = []
        if len(matches) == 0:
            doc._.sections.append((None, None, None, doc[0:]))
            return doc

        first_match = matches[0]
        section_spans = []
        if first_match[1] != 0:
            section_spans.append(Section(None, None, None, doc[0 : first_match[1]]))
        for i, match in enumerate(matches):
            (match_id, start, end, parent) = match
            section_header = doc[start:end]
            name = self.nlp.vocab.strings[match_id]
            # If this is the last match, it should include the rest of the doc
            if i == len(matches) - 1:
                if self.max_scope is None:
                    section_spans.append(
                        Section(name, section_header, parent, doc[start:])
                    )
                else:
                    scope_end = min(end + self.max_scope, doc[-1].i)
                    section_spans.append(
                        Section(name, section_header, parent, doc[start:scope_end])
                    )
            # Otherwise, go until the next section header
            else:
                next_match = matches[i + 1]
                _, next_start, _, _ = next_match
                if self.max_scope is None:
                    section_spans.append(
                        Section(name, section_header, parent, doc[start:next_start])
                    )
                else:
                    scope_end = min(end + self.max_scope, next_start)
                    section_spans.append(
                        Section(name, section_header, parent, doc[start:scope_end])
                    )

        # section_spans_with_parent = self.set_parent_sections(section_spans)

        # if there are no sections after required rules remove them, add one section over the entire document and exit
        # if len(section_spans_with_parent) == 0:
        #     doc._.sections.append((None, None, None, doc[0:]))
        #     return doc

        for section_tuple in section_spans:
            name, header, parent, section = section_tuple
            doc._.sections.append(section_tuple)
            for token in section:
                token._.section_span = section
                token._.section_title = name
                token._.section_header = header
                token._.section_parent = parent

        # If it is specified to add assertion attributes,
        # iterate through the entities in doc and add them
        if self.add_attrs is True:
            self.set_assertion_attributes(doc.ents)
        return doc

    def filter_start_lines(self, doc, matches):
        "Filter a list of matches to only contain spans where the start token is the beginning of a new line."
        return [
            m
            for m in matches
            if util.is_start_line(m[1], doc, self.newline_pattern)
        ]

    def filter_end_lines(self, doc, matches):
        "Filter a list of matches to only contain spans where the start token is followed by a new line."
        return [
            m
            for m in matches
            if util.is_end_line(m[2] - 1, doc, self.newline_pattern)
        ]


def prune_overlapping_matches(matches, strategy="longest"):
    if strategy != "longest":
        raise NotImplementedError()

    # Make a copy and sort
    unpruned = sorted(matches, key=lambda x: (x[1], x[2]))
    pruned = []
    num_matches = len(matches)
    if num_matches == 0:
        return matches
    curr_match = unpruned.pop(0)

    while True:
        if len(unpruned) == 0:
            pruned.append(curr_match)
            break
        next_match = unpruned.pop(0)

        # Check if they overlap
        if overlaps(curr_match, next_match):
            # Choose the larger span
            longer_span = max(
                curr_match, next_match, key=lambda x: (x[2] - x[1])
            )
            pruned.append(longer_span)
            if len(unpruned) == 0:
                break
            curr_match = unpruned.pop(0)
        else:
            pruned.append(curr_match)
            curr_match = next_match
    # Recursive base point
    if len(pruned) == num_matches:
        return pruned
    # Recursive function call
    else:
        return prune_overlapping_matches(pruned)


def overlaps(a, b):
    if _span_overlaps(a, b) or _span_overlaps(b, a):
        return True
    return False


def _span_overlaps(a, b):
    _, a_start, a_end = a
    _, b_start, b_end = b
    if a_start >= b_start and a_start < b_end:
        return True
    if a_end > b_start and a_end <= b_end:
        return True
    return False


def matches_to_spans(doc, matches, set_label=True):
    spans = []
    for (rule_id, start, end) in matches:
        if set_label:
            label = doc.vocab.strings[rule_id]
        else:
            label = None
        spans.append(Span(doc, start=start, end=end, label=label))
    return spans



from . import TargetMatcher
from spacy.tokens import Token

class ConceptTagger:
    name = "concept_tagger"

    def __init__(self, nlp, attr_name="concept_tag"):
        self.nlp = nlp
        self.attr_name = attr_name
        self.target_matcher = TargetMatcher(nlp, add_ents=False)
        self.rules = []

        # If the token attribute hasn't been set, add it now
        try:
            Token.set_extension(attr_name, default="")
        except:
            pass

    def add(self, rules):
        self.target_matcher.add(rules)
        for rule in rules:
            self.rules.append(rule)

    def __call__(self, doc):
        spans = self.target_matcher(doc)
        for span in spans:
            for token in span:
                setattr(token._, self.attr_name, span.label_)

        return doc



class PostprocessingPattern:

    def __init__(self, condition, success_value=True, condition_args=None):
        """A PostprocessingPattern defines a single condition to check against an entity.
        condition (function): A function to call on an entity. If the result of
            the function call equals success_value, then the pattern passes.
        success_value: The value which should be returned by condition(ent)
            in order for the pattern to pass. Default True.
        condition_args (function or None): Optional positional arguments to call
            with condition(ent, *condition_args). If None, will just call
            condition(ent). Default None.
        """
        self.condition = condition
        self.success_value = success_value
        self.condition_args = condition_args

    def __call__(self, ent):
        if self.condition_args is None:
            result = self.condition(ent)
        else:
            result = self.condition(ent, *self.condition_args)
        # print(result, self.success_value)
        return result == self.success_value
        # if result == self.success_value:
        #     return True
        # return False


class TargetRule:
    def __init__(self, literal, category, pattern=None, attributes=None, on_match=None):
        self.literal = literal
        self.category = category
        self.pattern = pattern
        self.attributes = attributes
        self.on_match = on_match
        self._rule_id = None
        
    def __repr__(self):
        return f"""TargetRule(literal="{self.literal}", category="{self.category}", pattern={self.pattern}, attributes={self.attributes}, on_match={self.on_match})"""



from collections import namedtuple
postprocess_pattern = namedtuple("PostProcessPattern", ["func", "attr", "check_value", "success_value"])

class Postprocessor:
    name = "postprocessor"

    def __init__(self, debug=False):
        self.rules = []
        self.debug = debug

    def add(self, rules):
        self.rules += rules

    def __call__(self, doc):
        # Iterate through the entities in reversed order
        for i in range(len(doc.ents)-1, -1, -1):
            ent = doc.ents[i]
            if self.debug:
                print(ent)
            for rule in self.rules:
                num_ents = len(doc.ents)
                rule(ent, i, self.debug)
                # Check if the entity was removed -if it was, skip to the next entity
                try:
                    if doc.ents[i] != ent:
                        break
                except IndexError:
                    break
            if self.debug:
                print()
        return doc



import re

# Filepath to default rules which are included in package
from os import path
from pathlib import Path

DEFAULT_RULES_FILEPATH = path.join(
    Path(__file__).resolve().parents[2],
    "resources",
    "section_patterns.jsonl",
)


class TextSectionizer:
    name = "text_sectionizer"

    def __init__(self, patterns="default"):
        self._patterns = []
        self._compiled_patterns = dict()
        self._section_titles = set()

        if patterns is not None:
            if patterns == "default":
                import os

                if not os.path.exists(DEFAULT_RULES_FILEPATH):
                    raise FileNotFoundError(
                        "The expected location of the default patterns file cannot be found. Please either "
                        "add patterns manually or add a jsonl file to the following location: ",
                        DEFAULT_RULES_FILEPATH,
                    )
                self.add(self.load_patterns_from_jsonl(DEFAULT_RULES_FILEPATH))
            # If a list, add each of the patterns in the list
            elif isinstance(patterns, list):
                self.add(patterns)
            elif isinstance(patterns, str):
                import os

                assert os.path.exists(patterns)
                self.add(self.load_patterns_from_jsonl(patterns))

    def add(self, patterns, cflags=None):
        """
        Add compiled regular expressions defined in patterns

        Positional arguments:
        - patterns --

        Keyword arguments:
        - cflags -- a list of regular expression compile flags
                 If cflags==None then cflags is set to [re.I]
        """

        def _mycomp(rex, flags=None):

            if flags is None:
                flags = [re.I]
            cflags = 0
            for f in flags:
                if isinstance(f, re.RegexFlag):
                    cflags = cflags | f
            return re.compile(rex, flags=cflags)

        for pattern_dict in patterns:
            name = pattern_dict["section_title"]
            pattern = pattern_dict["pattern"]
            if isinstance(pattern, str):
                self._compiled_patterns.setdefault(name, [])
                self._compiled_patterns[name].append(
                    _mycomp(pattern, flags=cflags)
                )
            else:
                # TODO: Change the default patterns
                # continue
                raise ValueError(
                    "Patterns added to the TextSectionizer must be strings",
                    pattern_dict,
                )
            self._patterns.append(pattern_dict)
            self._section_titles.add(name)

    @property
    def patterns(self):
        return self._patterns

    @property
    def section_titles(self):
        return self._section_titles

    @classmethod
    def load_patterns_from_jsonl(self, filepath):

        import json

        patterns = []
        with open(filepath) as f:
            for line in f:
                patterns.append(json.loads(line))

        return patterns

    def __call__(self, text):
        matches = []
        for (name, patterns) in self._compiled_patterns.items():
            for pattern in patterns:
                pattern_matches = list(pattern.finditer(text))
                for match in pattern_matches:
                    matches.append((name, match))

        if len(matches) == 0:
            return [(None, None, text)]

        matches = sorted(matches, key=lambda x: (x[1].start(), 0 - x[1].end()))
        matches = self._dedup_matches(matches)

        sections = []
        # If the first section doesn't start at the very beginning,
        # add an unknown section at the beginning
        if matches[0][1].start() != 0:
            sections.append((None, None, text[: matches[0][1].start()]))

        for i, (section_title, match) in enumerate(matches):
            section_header = match.group()
            # If this is the final section, it should include the rest of the text
            if i == len(matches) - 1:
                section_text = text[match.start() :]
                sections.append((section_title, section_header, section_text))
            # Otherwise, it will include all of the text up until the next section header
            else:
                next_match = matches[i + 1][1]
                section_text = text[match.start() : next_match.start()]
                sections.append((section_title, section_header, section_text))
        return sections

    def extract_sections(self, text):
        matches = []
        for name, sect_patterns in self.patterns.items():
            for pattern in sect_patterns:
                sect_matches = list(pattern.finditer(text))
                for match in sect_matches:
                    matches.append((name, match))
        if len(matches) == 0:
            return [(None, text)]

        matches = sorted(matches, key=lambda x: (x[1].start(), 0 - x[1].end()))
        matches = self._dedup_matches(matches)

        sections = []
        if matches[0][1].start() != 0:
            sections.append(("UNK", text[: matches[0][1].start()]))
        for i, (name, match) in enumerate(matches):
            if i == len(matches) - 1:
                sections.append((name, text[match.start() :]))
            else:
                next_match = matches[i + 1][1]
                sections.append(
                    (name, text[match.start() : next_match.start()])
                )

        return sections

    def _dedup_matches(self, matches):
        deduped = []
        # TODO: Make this smarter
        deduped.append(matches[0])
        for i, match in enumerate(matches[1:], start=1):
            if not self._overlaps(deduped[-1], match):
                deduped.append(match)
        return deduped

    def _overlaps(self, a, b):
        (_, a) = a
        (_, b) = b
        if a.start() <= b.start() < a.end():
            return True
        if b.start() <= a.start() < b.end():
            return True



class PreprocessingRule:

    def __init__(self, pattern, repl="", callback=None, desc=None):
        self.pattern = pattern
        self.repl = repl
        self.callback = callback
        self.desc = desc

    def __call__(self, text):
        """Apply a preprocessing rule. If the callback attribute of rule is None,
        then it will return a string using the rule pattern.sub method.
        If callback is not None, then then callback function will be executed using
        the the resulting match as an argument.
        """
        # If the rule just has a repl attribute,
        # Just return a simple re.sub
        if self.callback is None:
            return self.pattern.sub(self.repl, text)

        match = self.pattern.search(text)
        if match is None:
            return text
        return self.callback(match)



# from nlp_tools.utils import prune_overlapping_matches

from spacy.tokens import Token
from spacy.tokens import Span
from spacy.matcher import Matcher, PhraseMatcher

Token.set_extension("ignore", default=False, force=True)
Span.set_extension("target_attributes", default=None, force=True)
Span.set_extension("target_rule", default=None, force=True)
Span.set_extension("target_rule", default=None, force=True)


class TargetMatcher:
    name = "target_matcher"

    def __init__(self, nlp, add_ents=True):
        self.nlp = nlp
        self.add_ents = add_ents
        self._rules = list()
        self._rule_item_mapping = dict()
        self.labels = set()

        self.matcher = Matcher(self.nlp.vocab)
        self.phrase_matcher = PhraseMatcher(self.nlp.vocab, attr="LOWER")

    def add(self, rules):
        """Add a list of targetRules to the matcher."""
        i = len(self._rules)
        self._rules += rules

        for rule in rules:
            self.labels.add(rule.category)
            rule_id = f"{rule.category}_{i}"
            rule._rule_id = rule_id
            self._rule_item_mapping[rule_id] = rule
            if rule.pattern is not None:
                self.matcher.add(rule_id, [rule.pattern], on_match=rule.on_match)
            else:
                self.phrase_matcher.add(rule_id, rule.on_match, self.nlp.make_doc(rule.literal.lower()))
            i += 1

    def __call__(self, doc):
        matches = self.matcher(doc)
        matches += self.phrase_matcher(doc)
        matches = prune_overlapping_matches(matches)
        spans = []
        for (rule_id, start, end) in matches:
            rule = self._rule_item_mapping[self.nlp.vocab.strings[rule_id]]
            span = Span(doc, start=start, end=end, label=rule.category)
            span._.target_rule = rule
            if rule.attributes is not None:
                for (attribute, value) in rule.attributes.items():
                    try:
                        setattr(span._, attribute, value)
                    except AttributeError as e:
                        raise e
                span._.target_attributes = rule.attributes
            spans.append(span)
        if self.add_ents is True:
            for span in spans:
                try:
                    doc.ents += (span,)
                # spaCy will raise a value error if the token in span are already
                # part of an entity (ie., as part of an upstream component
                # In that case, let the existing span supercede this one
                except ValueError as e:
                    # raise e
                    pass
            return doc
        else:
            return spans



def prune_overlapping_matches(matches, strategy="longest"):
    if strategy != "longest":
        raise NotImplementedError()

    # Make a copy and sort
    unpruned = sorted(matches, key=lambda x: (x[1], x[2]))
    pruned = []
    num_matches = len(matches)
    if num_matches == 0:
        return matches
    curr_match = unpruned.pop(0)

    while True:
        if len(unpruned) == 0:
            pruned.append(curr_match)
            break
        next_match = unpruned.pop(0)

        # Check if they overlap
        if overlaps(curr_match, next_match):
            # Choose the larger span
            longer_span = max(curr_match, next_match, key=lambda x: (x[2] - x[1]))
            pruned.append(longer_span)
            if len(unpruned) == 0:
                break
            curr_match = unpruned.pop(0)
        else:
            pruned.append(curr_match)
            curr_match = next_match
    # Recursive base point
    if len(pruned) == num_matches:
        return pruned
    # Recursive function call
    else:
        return prune_overlapping_matches(pruned)

def overlaps(a, b):
    if _span_overlaps(a, b) or _span_overlaps(b, a):
        return True
    return False

def _span_overlaps(a, b):
    _, a_start, a_end = a
    _, b_start, b_end = b
    if a_start >= b_start and a_start < b_end:
        return True
    if a_end > b_start and a_end <= b_end:
        return True
    return False

def matches_to_spans(doc, matches, set_label=True):
    spans = []
    for (rule_id, start, end) in matches:
        if set_label:
            label = doc.vocab.strings[rule_id]
        else:
            label = None
        spans.append(Span(doc, start=start, end=end, label=label))
    return spans



class Preprocessor:
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer
        self.rules = []

    def add(self, rules):
        self.rules += rules

    def __call__(self, text):
        for rule in self.rules:
            text = rule(text)

        return self.tokenizer(text)








import pytest
import spacy
from spacy.tokens import Span, Doc

from medspacy.context import ConTextItem, ConTextComponent
from medspacy.context.tag_object import TagObject

nlp = spacy.load("en_core_web_sm")


class TestTagObject:
    def create_objects(self):
        doc = nlp("family history of breast cancer but no diabetes. She has afib.")
        item = ConTextItem("family history of", "FAMILY_HISTORY", rule="FORWARD")
        tag_object = TagObject(item, 0, 3, doc)
        return doc, item, tag_object

    def create_target_type_examples(self):
        doc = nlp("no history of travel to Puerto Rico pneumonia")
        span1 = Span(doc, start=5, end=7, label="TRAVEL")
        span2 = Span(doc, start=7, end=8, label="CONDITION")
        doc.ents = (span1, span2)
        return doc

    def create_num_target_examples(self):
        doc = nlp("Pt with diabetes, pneumonia vs COPD")
        spans = [
            Span(doc, 2, 3, "CONDITION"),
            Span(doc, 4, 5, "CONDITION"),
            Span(doc, 6, 7, "CONDITION"),
        ]
        doc.ents = spans
        return doc

    def create_num_target_examples_no_sent(self):
        doc = nlp.tokenizer("Pt with diabetes, pneumonia vs COPD")
        spans = [
            Span(doc, 2, 3, "CONDITION"),
            Span(doc, 4, 5, "CONDITION"),
            Span(doc, 6, 7, "CONDITION"),
        ]
        doc.ents = spans
        return doc

    def test_init(self):
        assert self.create_objects()

    def test_span(self):
        doc, item, tag_object = self.create_objects()
        assert tag_object.span == doc[0:3]

    def test_set_span_fails(self):
        doc, item, tag_object = self.create_objects()
        with pytest.raises(AttributeError):
            tag_object.span = "Can't do this!"

    def test_rule(self):
        doc, item, tag_object = self.create_objects()
        assert tag_object.rule == "FORWARD"

    def test_category(self):
        doc, item, tag_object = self.create_objects()
        assert tag_object.category == "FAMILY_HISTORY"

    def test_default_scope(self):
        """Test that the scope goes from the end of the modifier phrase
        to the end of the sentence.
        """
        doc, item, tag_object = self.create_objects()
        assert tag_object.scope == doc[3:-4]

    def test_limit_scope(self):
        """Test that a 'TERMINATE' TagObject limits the scope of the tag object"""
        doc, item, tag_object = self.create_objects()
        item2 = ConTextItem("but", "TERMINATE", "TERMINATE")
        tag_object2 = TagObject(item2, 2, 4, doc)
        assert tag_object.limit_scope(tag_object2)

    def test_limit_scope2(self):
        doc, item, tag_object = self.create_objects()
        item2 = ConTextItem("but", "TERMINATE", "TERMINATE")
        tag_object2 = TagObject(item2, 2, 4, doc)
        assert not tag_object2.limit_scope(tag_object)

    def test_limit_scope3(self):
        """Test that two modifiers of the same type limit the scope of the first modifier."""
        doc = nlp("no evidence of CHF, neg for pneumonia")
        item = ConTextItem("no evidence of", "DEFINITE_NEGATED_EXISTENCE", "FORWARD")
        item2 = ConTextItem("neg for", "DEFINITE_NEGATED_EXISTENCE", "FORWARD")
        tag_object = TagObject(item, 0, 3, doc)
        tag_object2 = TagObject(item2, 5, 7, doc)
        assert tag_object.limit_scope(tag_object2)

    def test_terminate_limit_scope_custom(self):
        """Test that a modifier will be explicitly terminated by a modifier with a category
        in terminated_by."""
        doc = nlp("negative for flu, positive for pneumonia.")
        item = ConTextItem("negative for", "NEGATED_EXISTENCE", rule="FORWARD", terminated_by={"POSITIVE_EXISTENCE"})
        item2 = ConTextItem("positive for", "POSITIVE_EXISTENCE", rule="FORWARD")
        tag_object = TagObject(item, 0, 2, doc)
        tag_object2 = TagObject(item2, 4, 6, doc)
        assert tag_object.limit_scope(tag_object2)

    def test_terminate_limit_scope_custom2(self):
        """Test that a modifier will be explicitly terminated by a modifier with a category
        in terminated_by."""
        doc = nlp("flu is negative, pneumonia is positive.")
        item = ConTextItem("negative", "NEGATED_EXISTENCE", rule="BACKWARD")
        item2 = ConTextItem("positive", "POSITIVE_EXISTENCE", rule="BACKWARD", terminated_by={"NEGATED_EXISTENCE"})
        tag_object = TagObject(item, 2, 3, doc)
        tag_object2 = TagObject(item2, 6, 7, doc)
        assert tag_object2.limit_scope(tag_object)

    def test_terminate_limit_scope_backward(self):
        """Test that a 'TERMINATE' modifier will limit the scope of a 'BACKWARD' modifier.
        """
        doc = nlp("Pt has chf but pneumonia is ruled out")
        item = ConTextItem("is ruled out", "NEGATED_EXISTENCE", "BACKWARD")
        tag_object = TagObject(item, 6, 8, doc)

        item2 = ConTextItem("but", "TERMINATE", "TERMINATE")
        tag_object2 = TagObject(item2, 3, 4, doc)
        assert tag_object.limit_scope(tag_object2)

    def test_terminate_stops_forward_modifier(self):
        context = ConTextComponent(nlp, rules=None)

        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", "FORWARD")
        item2 = ConTextItem("but", "TERMINATE", "TERMINATE")
        context.add([item, item2])
        doc = nlp("No evidence of chf but she has pneumonia.")
        doc.ents = (Span(doc, 3, 4, "PROBLEM"), Span(doc, 7, 8, "PROBLEM"))
        context(doc)
        chf, pneumonia = doc.ents
        assert len(chf._.modifiers) > 0
        assert len(pneumonia._.modifiers) == 0

    def test_terminate_stops_backward_modifier(self):
        context = ConTextComponent(nlp, rules=None)

        item = ConTextItem("is ruled out", "NEGATED_EXISTENCE", "BACKWARD")
        item2 = ConTextItem("but", "CONJ", "TERMINATE")
        context.add([item, item2])
        doc = nlp("Pt has chf but pneumonia is ruled out")
        doc.ents = (Span(doc, 2, 3, "PROBLEM"), Span(doc, 4, 5, "PROBLEM"))
        context(doc)
        chf, pneumonia = doc.ents
        assert len(chf._.modifiers) == 0
        assert len(pneumonia._.modifiers) > 0

    def test_no_custom_terminate_stops_forward_modifier(self):
        doc = nlp("negative for flu, positive for pneumonia.")
        context = ConTextComponent(nlp, rules=None)

        item = ConTextItem("negative for", "NEGATED_EXISTENCE", rule="FORWARD", terminated_by=None)
        item2 = ConTextItem("positive for", "POSITIVE_EXISTENCE", rule="FORWARD")
        context.add([item, item2])
        doc.ents = (Span(doc, 2, 3, "PROBLEM"), Span(doc, 6, 7))
        flu, pneumonia = doc.ents
        context(doc)
        assert len(flu._.modifiers) == 1
        assert len(pneumonia._.modifiers) == 2

    def test_custom_terminate_stops_forward_modifier(self):
        doc = nlp("negative for flu, positive for pneumonia.")
        context = ConTextComponent(nlp, rules=None)

        item = ConTextItem("negative for", "NEGATED_EXISTENCE", rule="FORWARD", terminated_by={"POSITIVE_EXISTENCE"})
        item2 = ConTextItem("positive for", "POSITIVE_EXISTENCE", rule="FORWARD")
        context.add([item, item2])
        doc.ents = (Span(doc, 2, 3, "PROBLEM"), Span(doc, 6, 7))
        flu, pneumonia = doc.ents
        context(doc)
        assert len(flu._.modifiers) == 1
        assert len(pneumonia._.modifiers) == 1


    def test_no_limit_scope_same_category_different_allowed_types(self):
        """Test that a two TagObjects of the same type but with different
         allowed types does not limits the scope of the tag object.
         """
        doc = nlp("no history of travel to Puerto Rico, neg for pneumonia")

        item = ConTextItem(
            "no history of",
            "DEFINITE_NEGATED_EXISTENCE",
            "FORWARD",
            allowed_types={"TRAVEL"},
        )
        item2 = ConTextItem(
            "neg for",
            "DEFINITE_NEGATED_EXISTENCE",
            "FORWARD",
            allowed_types={"CONDITION"},
        )
        tag_object = TagObject(item, 0, 3, doc)
        tag_object2 = TagObject(item2, 8, 10, doc)
        assert not tag_object.limit_scope(tag_object2)

    def test_set_scope_fails_no_sentences(self):
        """Test that setting the scope fails if sentence boundaries haven't been set."""
        doc = nlp.tokenizer("family history of breast cancer but no diabetes. She has afib.")
        item = ConTextItem("family history of", "FAMILY_HISTORY", rule="FORWARD")
        with pytest.raises(ValueError) as exception_info:
            # This should fail because doc.sents are None
            TagObject(item, 0, 3, doc)
        exception_info.match(
            "ConText failed because sentence boundaries have not been set"
        )

    def test_set_scope_context_window_no_sentences(self):
        """Test that setting the scope succeeds if sentence boundaries haven't been set but _use_context_window is True."""
        doc = nlp.tokenizer("family history of breast cancer but no diabetes. She has afib.")
        item = ConTextItem("family history of", "FAMILY_HISTORY", rule="FORWARD", max_scope=2)
        tag_object = TagObject(item, 0, 3, doc, _use_context_window=True)
        assert tag_object.scope == doc[3:5]

    def test_update_scope(self):
        doc, item, tag_object = self.create_objects()
        tag_object.update_scope(doc[3:5])

    def test_modifies(self):
        """Test that the TagObject modifies a target in its scope"""
        doc, item, tag_object = self.create_objects()
        assert tag_object.modifies(doc[3:5])

    def test_not_modifies(self):
        """Test that the TagObject does not modify a target outside of its scope"""
        doc, item, tag_object = self.create_objects()
        assert not tag_object.modifies(doc[-2:])

    def test_context_item(self):
        doc, item, tag_object = self.create_objects()
        assert tag_object.context_item is item

    def test_allowed_types(self):
        """Test that specifying allowed_types will only modify that target type."""
        doc = self.create_target_type_examples()
        item = ConTextItem(
            "no history of travel to",
            category="DEFINITE_NEGATED_EXISTENCE",
            rule="FORWARD",
            allowed_types={"TRAVEL"},
        )
        tag_object = TagObject(item, 0, 5, doc)
        tag_object.set_scope()
        travel, condition = doc.ents  # "puerto rico", "pneumonia"
        assert tag_object.modifies(travel) is True
        assert tag_object.modifies(condition) is False

    def test_excluded_types(self):
        """Test that specifying excluded_types will not modify that target type."""
        doc = self.create_target_type_examples()
        item = ConTextItem(
            "no history of travel to",
            category="DEFINITE_NEGATED_EXISTENCE",
            rule="FORWARD",
            excluded_types={"CONDITION"},
        )
        tag_object = TagObject(item, 0, 5, doc)
        tag_object.set_scope()
        travel, condition = doc.ents  # "puerto rico", "pneumonia"
        assert tag_object.modifies(travel) is True
        assert tag_object.modifies(condition) is False

    def test_no_types(self):
        """Test that not specifying allowed_types or excluded_types will modify all targets."""
        doc = self.create_target_type_examples()
        item = ConTextItem(
            "no history of travel to",
            category="DEFINITE_NEGATED_EXISTENCE",
            rule="FORWARD",
        )
        tag_object = TagObject(item, 0, 5, doc)
        tag_object.set_scope()
        travel, condition = doc.ents  # "puerto rico", "pneumonia"
        assert tag_object.modifies(travel) is True
        assert tag_object.modifies(condition) is True

    def test_max_targets_less_than_targets(self):
        """Check that if max_targets is not None it will reduce the targets
        to the two closest ents.
        """
        doc = self.create_num_target_examples()
        assert len(doc.ents) == 3
        item = ConTextItem(
            "vs", category="UNCERTAIN", rule="BIDIRECTIONAL", max_targets=2
        )
        # Set "vs" to be the modifier
        tag_object = TagObject(item, 5, 6, doc)
        for target in doc.ents:
            tag_object.modify(target)
        assert tag_object.num_targets == 3

        tag_object.reduce_targets()
        assert tag_object.num_targets == 2
        for target in tag_object._targets:
            assert target.lower_ in ("pneumonia", "copd")

    def test_max_targets_equal_to_targets(self):
        """Check that if max_targets is not None it will reduce the targets
        to the two closest ents.
        """
        doc = self.create_num_target_examples()
        assert len(doc.ents) == 3
        item = ConTextItem(
            "vs", category="UNCERTAIN", rule="BIDIRECTIONAL", max_targets=3
        )
        # Set "vs" to be the modifier
        tag_object = TagObject(item, 5, 6, doc)
        for target in doc.ents:
            tag_object.modify(target)
        assert tag_object.num_targets == 3

        tag_object.reduce_targets()
        assert tag_object.num_targets == 3

    def test_max_targets_none(self):
        """Check that if max_targets is None it will not reduce the targets
        to the two closest ents.
        """
        doc = self.create_num_target_examples()
        assert len(doc.ents) == 3
        item = ConTextItem(
            "vs", category="UNCERTAIN", rule="BIDIRECTIONAL", max_targets=None
        )
        # Set "vs" to be the modifier
        tag_object = TagObject(item, 5, 6, doc)
        for target in doc.ents:
            tag_object.modify(target)
        assert tag_object.num_targets == 3

        tag_object.reduce_targets()
        assert tag_object.num_targets == 3

    def test_max_scope(self):
        """Test that if max_scope is not None it will reduce the range
        of text which is modified.
        """
        doc = self.create_num_target_examples()
        assert len(doc.ents) == 3
        item = ConTextItem(
            "vs", category="UNCERTAIN", rule="BIDIRECTIONAL", max_scope=1
        )
        tag_object = TagObject(item, 5, 6, doc)

        for target in doc.ents:
            if tag_object.modifies(target):
                tag_object.modify(target)
        assert tag_object.num_targets == 2
        for target in tag_object._targets:
            assert target.lower_ in ("pneumonia", "copd")

    def test_max_scope_none(self):
        """Test that if max_scope is not None it will reduce the range
        of text which is modified.
        """
        doc = self.create_num_target_examples()
        assert len(doc.ents) == 3
        item = ConTextItem(
            "vs", category="UNCERTAIN", rule="BIDIRECTIONAL", max_scope=None
        )
        tag_object = TagObject(item, 5, 6, doc)

        for target in doc.ents:
            if tag_object.modifies(target):
                tag_object.modify(target)
        assert tag_object.num_targets == 3



    def test_overlapping_target(self):
        """Test that a modifier will not modify a target if it is
        in the same span as the modifier.
        """
        doc = nlp("Pt presents for r/o of pneumonia.")
        item = ConTextItem("r/o", "UNCERTAIN", rule="BIDIRECTIONAL")
        tag_object = TagObject(item, 3, 4, doc)
        target = Span(doc, 3, 4, "TEST")

        assert tag_object.modifies(target) is False

    def test_on_modifies_true(self):
        def on_modifies(target, modifier, span_between):
            return True

        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", on_modifies=on_modifies)
        doc = nlp("There is no evidence of pneumonia or chf.")
        doc.ents = (doc[5:6], doc[7:8])
        tag = TagObject(item, 2, 5, doc)

        assert tag.modifies(doc.ents[0]) is True

    def test_on_modifies_false(self):
        def on_modifies(target, modifier, span_between):
            return False

        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", on_modifies=on_modifies)
        doc = nlp("There is no evidence of pneumonia or chf.")
        doc.ents = (doc[5:6], doc[7:8])
        tag = TagObject(item, 2, 5, doc)

        assert tag.modifies(doc.ents[0]) is False


    def test_on_modifies_arg_types(self):
        def check_arg_types(target, modifier, span_between):
            for arg in (target, modifier, span_between):
                if not isinstance(arg, spacy.tokens.Span):
                    return False
            return True

        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", on_modifies=check_arg_types)
        doc = nlp("There is no evidence of pneumonia or chf.")
        doc.ents = (doc[5:6], doc[7:8])
        tag = TagObject(item, 2, 5, doc)

        assert tag.modifies(doc.ents[0]) is True

    def test_on_modifies_arg_values(self):
        def check_arg_types(target, modifier, span_between):
            if target.lower_ != "chf":
                return False
            if modifier.lower_ != "no evidence of":
                return False
            if span_between.lower_ != "pneumonia or":
                return False
            return True

        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", on_modifies=check_arg_types)
        doc = nlp("There is no evidence of pneumonia or chf.")
        doc.ents = (doc[5:6], doc[7:8])
        tag = TagObject(item, 2, 5, doc)

        assert tag.modifies(doc.ents[1]) is True



"""The ConTextComponent definiton."""
from os import path

# Filepath to default rules which are included in package
from pathlib import Path

from spacy.matcher import Matcher, PhraseMatcher
from spacy.tokens import Doc, Span

from .tag_object import TagObject
from .context_graph import ConTextGraph
from .context_item import ConTextItem

#
DEFAULT_ATTRS = {
    "NEGATED_EXISTENCE": {"is_negated": True},
    "POSSIBLE_EXISTENCE": {"is_uncertain": True},
    "HISTORICAL": {"is_historical": True},
    "HYPOTHETICAL": {"is_hypothetical": True},
    "FAMILY": {"is_family": True},
}

DEFAULT_RULES_FILEPATH = path.join(
    Path(__file__).resolve().parents[2], "resources", "context_rules.json"
)


class ConTextComponent:
    """The ConTextComponent for spaCy processing."""

    name = "context"

    def __init__(
        self,
        nlp,
        targets="ents",
        add_attrs=True,
        phrase_matcher_attr="LOWER",
        rules="default",
        rule_list=None,
        allowed_types=None,
        excluded_types=None,
        use_context_window=False,
        max_scope=None,
        max_targets=None,
        terminations=None,
        prune=True,
        remove_overlapping_modifiers=False,
    ):

        """Create a new ConTextComponent algorithm.

        This component matches modifiers in a Doc,
        defines their scope, and identifies edges between targets and modifiers.
        Sets two spaCy extensions:
            - Span._.modifiers: a list of TagObject objects which modify a target Span
            - Doc._.context_graph: a ConText graph object which contains the targets,
                modifiers, and edges between them.

        Args:
            nlp: a spaCy NLP model
            targets: the attribute of Doc which contains targets.
                Default is "ents", in which case it will use the standard Doc.ents attribute.
                Otherwise will look for a custom attribute in Doc._.{targets}
            add_attrs: Whether or not to add the additional spaCy Span attributes (ie., Span._.x)
                defining assertion on the targets. By default, these are:
                - is_negated: True if a target is modified by 'NEGATED_EXISTENCE', default False
                - is_uncertain: True if a target is modified by 'POSSIBLE_EXISTENCE', default False
                - is_historical: True if a target is modified by 'HISTORICAL', default False
                - is_hypothetical: True if a target is modified by 'HYPOTHETICAL', default False
                - is_family: True if a target is modified by 'FAMILY', default False
                In the future, these should be made customizable.
            phrase_matcher_attr: The token attribute to be used by the underlying PhraseMatcher.
                If "LOWER", then the matching of modifiers with a "literal" string will be
                case-insensitive. If "TEXT" or "ORTH", it will be case-sensitive.
                Default "LOWER'.
            prune: Whether or not to prune modifiers which are substrings of another modifier.
                For example, if "no history of" and "history of" are both ConTextItems, both will match
                the text "no history of afib", but only "no history of" should modify afib.
                If True, will drop shorter substrings completely.
                Default True.
            remove_overlapping_modifiers: Whether or not to remove any matched modifiers which overlap
                with target entities. If False, any overlapping modifiers will not modify the overlapping
                entity but will still modify any other targets in its scope.
                Default False.
            rules: Which rules to load on initialization. Default is 'default'.
                - 'default': Load the default set of rules provided with cyConText
                - 'other': Load a custom set of rules, please also set rule_list with a file path or list.
                - None: Load no rules.
            rule_list: The location of rules in json format or a list of ContextItems. Default
                is None.
            allowed_types (set or None): A set of target labels to allow a ConTextItem to modify.
                If None, will apply to any type not specifically excluded in excluded_types.
                Only one of allowed_types and excluded_types can be used. An error will be thrown
                if both or not None.
                If this attribute is also defined in the ConTextItem, it will keep that value.
                Otherwise it will inherit this value.
            excluded_types (set or None): A set of target labels which this modifier cannot modify.
                If None, will apply to all target types unless allowed_types is not None.
                If this attribute is also defined in the ConTextItem, it will keep that value.
                Otherwise it will inherit this value.
            max_targets (int or None): The maximum number of targets which a modifier can modify.
                If None, will modify all targets in its scope.
                If this attribute is also defined in the ConTextItem, it will keep that value.
                Otherwise it will inherit this value.
            use_context_window (bool): Whether to use a specified range around a target to check
                for modifiers rather than split sentence boundaries. This can be useful
                for quicker processing by skipping sentence splitting or errors caused by poorly
                defined sentence boundaries. If True, max_scope must be an integer greater than 0.
            max_scope (int or None): A number to explicitly limit the size of the modifier's scope
                If this attribute is also defined in the ConTextItem, it will keep that value.
                Otherwise it will inherit this value.
            terminations (dict or None): Optional mapping between different categories which will
                cause one modifier type to be 'terminated' by another type. For example, if given
                a mapping:
                    {"POSITIVE_EXISTENCE": {"NEGATED_EXISTENCE", "UNCERTAIN"},
                    "NEGATED_EXISTENCE": {"FUTURE"},
                    }
                all modifiers of type "POSITIVE_EXISTENCE" will be terminated by "NEGATED_EXISTENCE" or "UNCERTAIN"
                modifiers, and all "NEGATED_EXISTENCE" modifiers will be terminated by "FUTURE".
                This can also be defined for specific ConTextItems in the `terminated_by` attribute.


        Returns:
            context: a ConTextComponent

        Raises:
            ValueError: if one of the parameters is incorrectly formatted.
        """

        self.nlp = nlp
        if targets != "ents":
            raise NotImplementedError()
        self._target_attr = targets
        self.prune = prune
        self.remove_overlapping_modifiers = remove_overlapping_modifiers

        self._item_data = []
        self._i = 0
        self._categories = set()

        # _modifier_item_mapping: A mapping from spaCy Matcher match_ids to ConTextItem
        # This allows us to use spaCy Matchers while still linking back to the ConTextItem
        # To get the rule and category
        self._modifier_item_mapping = dict()
        self.phrase_matcher = PhraseMatcher(
            nlp.vocab, attr=phrase_matcher_attr, validate=True
        )  # TODO: match on custom attributes
        self.matcher = Matcher(nlp.vocab, validate=True)

        self.register_graph_attributes()
        if add_attrs is False:
            self.add_attrs = False
        elif add_attrs is True:
            self.add_attrs = True
            self.context_attributes_mapping = DEFAULT_ATTRS
            self.register_default_attributes()
        elif isinstance(add_attrs, dict):
            # Check that each of the attributes being added has been set
            for modifier in add_attrs.keys():
                attr_dict = add_attrs[modifier]
                for attr_name, attr_value in attr_dict.items():
                    if not Span.has_extension(attr_name):
                        raise ValueError(
                            "Custom extension {0} has not been set. Call Span.set_extension.".format(
                                attr_name
                            )
                        )

            self.add_attrs = True
            self.context_attributes_mapping = add_attrs

        else:
            raise ValueError(
                "add_attrs must be either True (default), False, or a dictionary, not {0}".format(
                    add_attrs
                )
            )
        if use_context_window is True:
            if not isinstance(max_scope, int) or max_scope < 1:
                raise ValueError(
                    "If 'use_context_window' is True, 'max_scope' must be an integer greater 1, "
                    "not {0}".format(max_scope)
                )
        self.use_context_window = use_context_window
        if max_scope is not None and (
            not isinstance(max_scope, int) or max_scope < 1
        ):
            raise ValueError(
                "'max_scope' must be None or an integer greater 1, "
                "not {0}".format(max_scope)
            )
        self.max_scope = max_scope

        self.allowed_types = allowed_types
        self.excluded_types = excluded_types
        self.max_targets = max_targets

        if terminations is None:
            terminations = dict()
        self.terminations = {k.upper(): v for (k, v) in terminations.items()}

        if rules == "default":

            item_data = ConTextItem.from_json(DEFAULT_RULES_FILEPATH)
            self.add(item_data)

        elif rules == "other":
            # use custom rules
            if isinstance(rule_list, str):
                # if rules_list is a string, then it must be a path to a json
                if "yaml" in rule_list or "yml" in rule_list:
                    try:
                        rule_list = ConTextItem.from_yaml(rule_list)
                    except:
                        raise ValueError(
                            "rule list {0} could not be read".format(rule_list)
                        )
                elif path.exists(rule_list):
                    item_data = ConTextItem.from_json(rule_list)
                    self.add(item_data)
                else:
                    raise ValueError(
                        "rule_list must be a valid path. Currently is: {0}".format(
                            rule_list
                        )
                    )

            elif isinstance(rule_list, list):
                # otherwise it is a list of contextitems
                if not rule_list:
                    raise ValueError("rule_list must not be empty.")
                for item in rule_list:
                    # check that all items are contextitems
                    if not isinstance(item, ConTextItem):
                        raise ValueError(
                            "rule_list must contain only ContextItems. Currently contains: {0}".format(
                                type(item)
                            )
                        )
                self.add(rule_list)

            else:
                raise ValueError(
                    "rule_list must be a valid path or list of ContextItems. Currenty is: {0}".format(
                        type(rule_list)
                    )
                )

        elif not rules:
            # otherwise leave the list empty.
            # do nothing
            self._item_data = []

        else:
            # loading from json path or list is possible later
            raise ValueError(
                "rules must either be 'default' (default), 'other' or None."
            )

    @property
    def item_data(self):
        """Returns list of ConTextItems"""
        return self._item_data

    @property
    def categories(self):
        """Returns list of categories from ConTextItems"""
        return self._categories

    def add(self, item_data):
        """Add a list of ConTextItem items to ConText.

        Args:
            item_data: a list of ConTextItems to add.

        Raises:
            TypeError: if item_data contains an object that is not a ConTextItem.
        """
        try:
            self._item_data += item_data
        except TypeError:
            raise TypeError(
                "item_data must be a list of ConText items. If you're just passing in a single ConText Item, "
                "make sure to wrap the item in a list: `context.add([item])`"
            )

        for item in item_data:

            # UID is the hash which we'll use to retrieve the ConTextItem from a spaCy match
            # And will be a key in self._modifier_item_mapping
            uid = self.nlp.vocab.strings[str(self._i)]
            # If no pattern is defined,
            # match on the literal phrase.
            if item.pattern is None:
                self.phrase_matcher.add(
                    str(self._i),
                    [self.nlp.make_doc(item.literal)],
                    on_match=item.on_match,
                )
            else:

                self.matcher.add(
                    str(self._i), [item.pattern], on_match=item.on_match
                )
            self._modifier_item_mapping[uid] = item
            self._i += 1
            self._categories.add(item.category)

            # If global attributes like allowed_types and max_scope are defined,
            # check if the ConTextItem has them defined. If not, set to the global
            for attr in (
                "allowed_types",
                "excluded_types",
                "max_scope",
                "max_targets",
            ):
                value = getattr(self, attr)
                if value is None:  # No global value set
                    continue
                if (
                    getattr(item, attr) is None
                ):  # If the item itself has it defined, don't override
                    setattr(item, attr, value)

            # Check custom termination points
            if item.category.upper() in self.terminations:
                for other_modifier in self.terminations[item.category.upper()]:
                    item.terminated_by.add(other_modifier.upper())

    def register_default_attributes(self):
        """Register the default values for the Span attributes defined in DEFAULT_ATTRS."""
        for attr_name in [
            "is_negated",
            "is_uncertain",
            "is_historical",
            "is_hypothetical",
            "is_family",
        ]:
            try:
                Span.set_extension(attr_name, default=False)
            except ValueError:  # Extension already set
                pass

    def register_graph_attributes(self):
        """Register spaCy container custom attribute extensions.

        By default will register Span._.modifiers and Doc._.context_graph.

        If self.add_attrs is True, will add additional attributes to span
            as defined in DEFAULT_ATTRS:
            - is_negated
            - is_historical
            - is_experiencer
        """
        Span.set_extension("modifiers", default=(), force=True)
        Doc.set_extension("context_graph", default=None, force=True)

    def set_context_attributes(self, edges):
        """Add Span-level attributes to targets with modifiers.

        Args:
            edges: the edges to modify

        """

        for (target, modifier) in edges:
            if modifier.category in self.context_attributes_mapping:
                attr_dict = self.context_attributes_mapping[modifier.category]
                for attr_name, attr_value in attr_dict.items():
                    setattr(target._, attr_name, attr_value)

    def __call__(self, doc):
        """Applies the ConText algorithm to a Doc.

        Args:
            doc: a spaCy Doc

        Returns:
            doc: a spaCy Doc
        """
        if self._target_attr == "ents":
            targets = doc.ents
        else:
            targets = getattr(doc._, self._target_attr)

        # Store data in ConTextGraph object
        # TODO: move some of this over to ConTextGraph
        context_graph = ConTextGraph(
            remove_overlapping_modifiers=self.remove_overlapping_modifiers
        )

        context_graph.targets = targets

        context_graph.modifiers = []

        matches = self.phrase_matcher(doc)
        matches += self.matcher(doc)

        # Sort matches
        matches = sorted(matches, key=lambda x: x[1])
        for (match_id, start, end) in matches:
            # Get the ConTextItem object defining this modifier
            item_data = self._modifier_item_mapping[match_id]
            tag_object = TagObject(
                item_data, start, end, doc, self.use_context_window
            )
            context_graph.modifiers.append(tag_object)

        if self.prune:
            context_graph.prune_modifiers()
        context_graph.update_scopes()
        context_graph.apply_modifiers()

        # Link targets to their modifiers
        for target, modifier in context_graph.edges:
            target._.modifiers += (modifier,)

        # If add_attrs is True, add is_negated, is_current, is_asserted to targets
        if self.add_attrs:
            self.set_context_attributes(context_graph.edges)

        doc._.context_graph = context_graph

        return doc



class ConTextGraph:
    def __init__(self, remove_overlapping_modifiers=False):
        self.targets = []
        self.modifiers = []
        self.edges = []
        self.remove_overlapping_modifiers = remove_overlapping_modifiers

    def update_scopes(self):
        """Update the scope of all TagObjects.

        For each modifier in a list of TagObjects, check against each other
        modifier to see if one of the modifiers should update the other. 
        This allows neighboring similar modifiers to extend each other's 
        scope and allows "terminate" modifiers to end a modifier's scope.

        Args:
            marked_modifiers: A list of TagObjects in a Doc.
        """
        for i in range(len(self.modifiers) - 1):
            modifier1 = self.modifiers[i]
            for j in range(i + 1, len(self.modifiers)):
                modifier2 = self.modifiers[j]
                # TODO: Add modifier -> modifier edges
                modifier1.limit_scope(modifier2)
                modifier2.limit_scope(modifier1)

    def apply_modifiers(self):
        """Checks each target/modifier pair. If modifier modifies target,
        create an edge between them.

        Args:
            marked_targets: A list of Spans
            marked_modifiers: A list of TagObjects

        RETURNS 
            edges: A list of tuples consisting of target/modifier pairs
        """
        if self.remove_overlapping_modifiers:
            for i in range(len(self.modifiers) - 1, -1, -1):
                modifier = self.modifiers[i]
                for target in self.targets:
                    if overlap_target_modifiers(target, modifier.span):
                        self.modifiers.pop(i)
                        break

        edges = []
        for target in self.targets:
            for modifier in self.modifiers:
                if modifier.modifies(target):
                    modifier.modify(target)

        # Now do a second pass and reduce the number of targets
        # for any modifiers with a max_targets int
        for modifier in self.modifiers:
            modifier.reduce_targets()
            for target in modifier._targets:
                edges.append((target, modifier))

        self.edges = edges

    def prune_modifiers(self):
        """Prune overlapping modifiers so that only the longest span is kept.

        For example, if "no" and "no evidence of" are both tagged as modifiers,
        only "no evidence of" will be kept.

        Additionally, this removes any modifiers which overlap with a target.
        For example, if doc.ents contains a span "does not know" and "not" is tagged by
        context as a modifier, "not" will be removed.

        # TODO: Consider only removing modifiers which are subspans.
        """

        unpruned = sorted(self.modifiers, key=lambda x: (x.end - x.end))
        if len(unpruned) > 0:
            rslt = self.prune_overlapping_modifiers(unpruned)
            self.modifiers = rslt

    def prune_overlapping_modifiers(self, modifiers):
        # Don't prune a single modifier
        if len(modifiers) == 1:
            return modifiers

        # Make a copy
        unpruned = list(modifiers)
        pruned = []
        num_mods = len(unpruned)
        curr_mod = unpruned.pop(0)

        while True:
            if len(unpruned) == 0:
                pruned.append(curr_mod)
                break
            # if len(unpruned) == 1:
            #     pruned.append(unpruned.pop(0))
            #     break
            next_mod = unpruned.pop(0)

            # Check if they overlap
            if curr_mod.overlaps(next_mod):
                # Choose the larger
                longer_span = max(curr_mod, next_mod, key=lambda x: (x.end - x.start))

                pruned.append(longer_span)
                if len(unpruned) == 0:
                    break
                curr_mod = unpruned.pop(0)
            else:
                pruned.append(curr_mod)
                curr_mod = next_mod
        # Recursion base point
        if len(pruned) == num_mods:
            return pruned
        return self.prune_overlapping_modifiers(pruned)

    def __repr__(self):
        return "<ConTextGraph> with {0} targets and {1} modifiers".format(
            len(self.targets), len(self.modifiers)
        )


def overlap_target_modifiers(span1, span2):
    """Checks whether two modifiers overlap.
        
    Args:
        span1: the first span
        span2: the second span
    """
    return _spans_overlap(span1, span2)


def _spans_overlap(span1, span2):
    return (span1.end > span2.start and span1.end <= span2.end) or (
        span1.start >= span2.start and span1.start < span2.end
    )



import pytest
import spacy

from medspacy.context import ConTextComponent
from medspacy.context import ConTextItem
from medspacy.context.tag_object import TagObject
from medspacy.context.context_graph import ConTextGraph
from spacy.tokens import Span
from medspacy.context.context_graph import overlap_target_modifiers

nlp = spacy.load("en_core_web_sm")


class TestConTextGraph:
    def context_graph(self):
        doc = nlp.tokenizer("There is no evidence of pneumonia but there is chf.")
        doc[0].is_sent_start = True
        for token in doc[1:]:
            token.is_sent_start = False
        item_data1 = ConTextItem(
            "no evidence of", "DEFINITE_NEGATED_EXISTENCE", "forward"
        )
        tag_object1 = TagObject(item_data1, 2, 5, doc)

        item_data2 = ConTextItem("evidence of", "DEFINITE_EXISTENCE", "forward")
        tag_object2 = TagObject(item_data2, 3, 5, doc)

        item_data3 = ConTextItem("but", "TERMINATE", "TERMINATE")
        tag_object3 = TagObject(item_data3, 6, 7, doc)

        graph = ConTextGraph()
        graph.modifiers = [tag_object1, tag_object2, tag_object3]
        return doc, graph

    def test_init(self):
        assert ConTextGraph()

    def test_apply_modifiers(self):
        doc, graph = self.context_graph()
        graph.targets = [doc[5:6]]  # "pneumonia"
        graph.apply_modifiers()
        assert len(graph.edges) == 2

    def test_prune_modifiers(self):
        doc, graph = self.context_graph()
        graph.targets = [doc[5:6]]  # "pneumonia"
        graph.prune_modifiers()
        assert len(graph.modifiers) == 2

    def test_update_scopes(self):
        doc, graph = self.context_graph()
        graph.targets = [doc[5:6]]  # "pneumonia"
        graph.apply_modifiers()
        assert graph.modifiers[0].scope == doc[5:]
        graph.update_scopes()
        assert graph.modifiers[0].scope == doc[5:6]

    def test_remove_modifiers_overlap_target(self):
        """Test that a modifier which overlaps with a target is removed when set to True."""
        doc = nlp("The patient has heart failure.")
        doc.ents = (Span(doc, 3, 5, "CONDITION"),)
        context_item = ConTextItem("failure", "MODIFIER")
        tag_object = TagObject(context_item, 4, 5, doc)
        graph = ConTextGraph(remove_overlapping_modifiers=True)

        graph.modifiers = [tag_object]
        graph.targets = doc.ents
        graph.prune_modifiers()
        graph.apply_modifiers()

        assert overlap_target_modifiers(tag_object.span, doc.ents[0])
        assert len(graph.modifiers) == 0

    def test_not_remove_modifiers_overlap_target(self):
        """Test that a modifier which overlaps with a target is not pruned but does not modify itself."""
        doc = nlp("The patient has heart failure.")
        doc.ents = (Span(doc, 3, 5, "CONDITION"),)
        context_item = ConTextItem("failure", "MODIFIER")
        tag_object = TagObject(context_item, 4, 5, doc)
        graph = ConTextGraph(remove_overlapping_modifiers=False)

        graph.modifiers = [tag_object]
        graph.targets = doc.ents
        graph.prune_modifiers()
        graph.apply_modifiers()

        assert overlap_target_modifiers(tag_object.span, doc.ents[0])
        assert len(graph.modifiers) == 1





import pytest
import tempfile

tmpdirname = tempfile.TemporaryDirectory()

from medspacy.context import ConTextItem


class TestItemData:
    def test_instantiate1(self):
        literal = "no evidence of"
        category = "DEFINITE_NEGATED_EXISTENCE"
        rule = "forward"
        assert ConTextItem(literal, category, rule)

    def test_context_item_category_upper(self):
        """Test that a ConTextItem category is always upper"""
        literal = "no evidence of"
        category = "definite_negated_existence"
        rule = "forward"
        item = ConTextItem(literal, category, rule)
        assert item.category == "DEFINITE_NEGATED_EXISTENCE"

    def test_context_item_rule_upper(self):
        """Test that a ConTextItem rule is always upper"""
        literal = "no evidence of"
        category = "definite_negated_existence"
        rule = "forward"
        item = ConTextItem(literal, category, rule)
        assert item.rule == "FORWARD"

    def test_rule_value_error(self):
        """Test that ConTextItem raises a ValueError if an invalid rule is passed in."""
        literal = "no evidence of"
        category = "definite_negated_existence"
        rule = "asdf"
        with pytest.raises(ValueError):
            ConTextItem(literal, category, rule)

    def test_metadata(self):
        literal = "no evidence of"
        category = "definite_negated_existence"
        rule = "forward"
        meta = {"comment": "This is a comment."}
        item = ConTextItem(literal, category, rule, metadata=meta)
        assert item.metadata

    def test_from_dict(self):
        d = dict(
            literal="reason for examination", category="INDICATION", rule="FORWARD"
        )
        assert ConTextItem.from_dict(d)

    def test_from_dict_error(self):
        d = dict(
            literal="reason for examination",
            category="INDICATION",
            rule="FORWARD",
            invalid="this is an invalid key",
        )
        with pytest.raises(ValueError):
            ConTextItem.from_dict(d)

    def test_from_json(self, from_json_file):
        assert ConTextItem.from_json(from_json_file)

    def test_to_dict(self):
        literal = "no evidence of"
        category = "definite_negated_existence"
        rule = "forward"
        item = ConTextItem(literal, category, rule)
        assert isinstance(item.to_dict(), dict)

    def test_to_json(self):
        import json, os

        dname = os.path.join(tmpdirname.name, "test_modifiers.json")

        literal = "no evidence of"
        category = "definite_negated_existence"
        rule = "forward"
        item = ConTextItem(literal, category, rule)
        ConTextItem.to_json([item], dname)

        with open(dname) as f:
            data = json.load(f)
        assert "item_data" in data
        assert len(data["item_data"]) == 1
        item = data["item_data"][0]
        for key in ["literal", "category", "rule"]:
            assert key in item

        #os.remove("test_modifiers.json")

    def test_default_terminate(self):
        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", "FORWARD", terminated_by=None)
        assert item.terminated_by == set()

    def test_custom_terminate(self):
        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", "FORWARD", terminated_by={"POSITIVE_EXISTENCE"})
        assert item.terminated_by == {"POSITIVE_EXISTENCE"}


@pytest.fixture
def from_json_file():
    import json, os

    json_filepath = os.path.join(tmpdirname.name, "test_modifiers.json")

    item_data = [
        {
            "literal": "are ruled out",
            "category": "DEFINITE_NEGATED_EXISTENCE",
            "pattern": None,
            "rule": "backward",
        },
        {
            "literal": "is negative",
            "category": "DEFINITE_NEGATED_EXISTENCE",
            "pattern": [{"LEMMA": "be"}, {"LOWER": "negative"}],
            "rule": "backward",
        },
    ]

    # Save dicts to a temporary file
    with open(json_filepath, "w") as f:
        json.dump({"item_data": item_data}, f)

    yield json_filepath
    # os.remove(json_filepath)



"""This module contains functions to be used both as action and condition functions
for postprocessing patterns.
"""

# Condition functions

def is_negated(span):
    """Return True if a span is marked as negated by cycontext."""
    return span._.is_negated

def is_uncertain(span):
    """Return True if a span is marked as uncertain by cycontext."""
    return span._.is_uncertain

def is_historical(span):
    """Return True if a span is marked as historical by cycontext."""
    return span._.is_historical

def is_hypothetical(span):
    """Return True if a span is marked as hypothetical by cycontext."""
    return span._.is_hypothetical

def is_family(span):
    """Return True if a span is marked as family by cycontext."""
    return span._.is_family

def is_modified_by_category(span, category):
    """Returns True if a span is modified by a cycontext TagObject
    modifier with a certain category. Case insensitive.
    """
    for modifier in span._.modifiers:
        if modifier.category.upper() == category.upper():
            return True
    return False

def is_modified_by_text(span, target, regex=True):
    """Returns True if a span is modified by a cycontext TabObject
    modifier with a certain text.
    """
    for modifier in span._.modifiers:
        if span_contains(modifier.span, target, regex):
            return True
    return False

def is_preceded_by(ent, target, window=1):
    """Check if an entity is preceded by a target word within a certain window.
    Case-insensitive.
    If any phrases in target are more than one token long, this may not capture it
    if window is smaller than the number of tokens.
    ent (Span):  A spaCy Span
    target (str or iterable): Either a single string or iterable of strings.
        If an iterable, will return True if any of the strings are in the window
        preceding ent.
    """
    preceding_span = ent.doc[ent.start - window: ent.start]
    preceding_string = " ".join([token.lower_ for token in preceding_span])
    if isinstance(target, str):
        return target.lower() in preceding_string
    for string in target:
        if string.lower() in preceding_string:
            return True
    return False


def is_followed_by(ent, target, window=1):
    """Check if an entity is followed by a target word within a certain window.
    Case-insensitive.
    If any phrases in target are more than one token long, this may not capture it
    if window is smaller than the number of tokens.
    ent (Span):  A spaCy Span
    target (str or iterable): Either a single string or iterable of strings.
        If an iterable, will return True if any of the strings are in the window
        following ent.
    """
    following_span = ent.doc[ent.end: ent.end+window]
    following_string = " ".join([token.lower_ for token in following_span])
    if isinstance(target, str):
        return target.lower() in following_string
    for string in target:
        if string.lower() in following_string:
            return True
    return False

def span_contains(span, target, regex=True):
    """Return True if a Span object contains a target phrase.
    Case insensitive.
    span: A spaCy Span, such as an entity in doc.ents
    target: A target phrase or iterable of phrases to check in span.lower_.
    regex (bool): Whether to search the span using a regular expression rather than
        a literal string. Default True.
    """
    if regex is True:
        import re
        func = lambda x: re.search(x, span.lower_, re.IGNORECASE) is not None
    else:
        func = lambda x: x.lower() in span.lower_

    if isinstance(target, str):
        return func(target)

    # If it's an iterable, check if any of the strings are in sent
    for string in target:
        if func(string):
            return True
    return False

def ent_contains(ent, target, regex=True):
    """Check if an entity occurs in the same sentence as another span of text.
    ent (Span): A spaCy Span
    target (str or iterable): Either a single string or iterable of strings.
        If an iterable, will return True if any of the strings are a substring
        of ent.sent.
    Case insensitive.
    """
    return span_contains(ent, target, regex)


def sentence_contains(ent, target, regex=True):
    """Check if an entity occurs in the same sentence as another span of text.
    ent (Span): A spaCy Span
    target (str or iterable): Either a single string or iterable of strings.
        If an iterable, will return True if any of the strings are a substring
        of ent.sent.
    Case insensitive.
    """
    return span_contains(ent.sent, target, regex)

# Action funcs

def remove_ent(ent, i):
    """Remove an entity at position [i] from doc.ents."""
    ent.doc.ents = ent.doc.ents[:i] + ent.doc.ents[i+1:]

def set_label(ent, i, label):
    """Create a copy of the entity with a new label.
    WARNING: This is not fully safe, as spaCy does not allow modifying the label
    of a span. Instead this creates a new copy and attempts to copy existing
    attributes, but this is not totally reliable.
    """
    from spacy.tokens import Span
    new_ent = Span(ent.doc, ent.start, ent.end, label=label)
    # Copy any additional attributes
    # NOTE: This may not be complete and should be used with caution
    for (attr, values) in ent._.__dict__["_extensions"].items():
        setattr(new_ent._, attr, values[0])
    if len(ent.doc.ents) == 1:
        ent.doc.ents = (new_ent,)
    else:
        try:
            ent.doc.ents = ent.doc.ents[:i] + (new_ent,) + ent.doc.ents[i+1:]
        except ValueError: # Overlaps with another entity - debug later
            pass

def set_negated(ent, i, value=True):
    "Set the value of ent._.is_negated to value."
    ent._.is_negated = value

def set_uncertain(ent, i, value=True):
    "Set the value of ent._.is_uncertain to value."
    ent._.is_uncertain = value

def set_historical(ent, i, value=True):
    "Set the value of ent._.is_historical to value."
    ent._.is_historical = value

def set_hypothetical(ent, i, value=True):
    "Set the value of ent._.is_hypothetical to value."
    ent._.is_hypothetical = value

def set_family(ent, i, value=True):
    "Set the value of ent._.is_family to value."
    ent._.is_hypothetical = value



import spacy
from spacy.tokens import Span

from medspacy.context import ConTextComponent
from medspacy.context import ConTextItem

import pytest
import os

nlp = spacy.load("en_core_web_sm")


class TestConTextComponent:
    def test_initiate(self):
        assert ConTextComponent(nlp)

    def test_default_patterns(self):
        """Test that default rules are loaded"""
        context = ConTextComponent(nlp)
        assert context.item_data

    def test_empty_patterns(self):
        """Test that no rules are loaded"""
        context = ConTextComponent(nlp, rules=None)
        assert not context.item_data

    def test_custom_patterns_json(self):
        """Test that rules are loaded from a json"""
        context = ConTextComponent(
            nlp, rules="other", rule_list=os.path.abspath("resources/context_rules.json")
        )
        assert context.item_data

    def test_custom_patterns_list(self):
        """Test that rules are loaded from a list"""
        item = ConTextItem("evidence of", "DEFINITE_EXISTENCE", "forward")
        context = ConTextComponent(nlp, rules="other", rule_list=[item])
        assert context.item_data

    def test_bad_rules_arg(self):
        with pytest.raises(ValueError):
            ConTextComponent(nlp, rules="not valid")

    def test_bad_rule_list_path(self):
        with pytest.raises(ValueError):
            ConTextComponent(nlp, rules="other", rule_list="not a path")

    def test_bad_rule_list_empty(self):
        with pytest.raises(ValueError):
            ConTextComponent(nlp, rules="other", rule_list=[])

    def test_bad_rule_list_items(self):
        with pytest.raises(ValueError):
            ConTextComponent(nlp, rules="other", rule_list=["list of strings"])

    def test_call(self):
        doc = nlp("Pulmonary embolism has been ruled out.")
        context = ConTextComponent(nlp)
        doc = context(doc)
        assert isinstance(doc, spacy.tokens.doc.Doc)

    def test_registers_attributes(self):
        """Test that the default ConText attributes are set on ."""
        doc = nlp("There is consolidation.")
        doc.ents = (doc[-2:-1],)
        context = ConTextComponent(nlp)
        doc = context(doc)
        assert hasattr(doc._, "context_graph")
        assert hasattr(doc.ents[0]._, "modifiers")

    def test_registers_context_attributes(self):
        """Test that the additional attributes such as
        'is_negated' are registered on spaCy spans.
        """
        doc = nlp("This is a span.")
        context = ConTextComponent(nlp, add_attrs=True, rules=None)
        context(doc)
        span = doc[-2:]
        for attr_name in [
            "is_negated",
            "is_uncertain",
            "is_historical",
            "is_hypothetical",
            "is_family",
        ]:
            assert hasattr(span._, attr_name)

    def test_default_attribute_values(self):
        """Check that default Span attributes have False values without any modifiers."""
        doc = nlp("There is evidence of pneumonia.")
        context = ConTextComponent(nlp, add_attrs=True, rules=None)
        doc.ents = (doc[-2:-1],)
        context(doc)
        for attr_name in [
            "is_negated",
            "is_uncertain",
            "is_historical",
            "is_hypothetical",
            "is_family",
        ]:
            assert getattr(doc.ents[0]._, attr_name) is False

    def test_default_rules_match(self):
        context = ConTextComponent(nlp)
        matcher = context.matcher
        assert matcher(nlp("no evidence of"))

    def test_custom_rules_match(self):
        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", "forward")
        context = ConTextComponent(nlp, rules="other", rule_list=[item])
        matcher = context.phrase_matcher
        assert matcher(nlp("no evidence of"))

    def test_is_negated(self):
        doc = nlp("There is no evidence of pneumonia.")
        context = ConTextComponent(nlp, add_attrs=True, rules=None)
        item_data = [ConTextItem("no evidence of", "NEGATED_EXISTENCE", rule="forward")]
        context.add(item_data)
        doc.ents = (doc[-2:-1],)
        context(doc)

        assert doc.ents[0]._.is_negated is True

    def test_is_historical(self):
        doc = nlp("History of pneumonia.")
        context = ConTextComponent(nlp, add_attrs=True, rules=None)
        item_data = [ConTextItem("history of", "HISTORICAL", rule="forward")]
        context.add(item_data)
        doc.ents = (doc[-2:-1],)
        context(doc)

        assert doc.ents[0]._.is_historical is True

    def test_is_family(self):
        doc = nlp("Family history of breast cancer.")
        context = ConTextComponent(nlp, add_attrs=True, rules=None)
        item_data = [ConTextItem("family history of", "FAMILY", rule="forward")]
        context.add(item_data)
        doc.ents = (doc[-3:-1],)
        context(doc)

        assert doc.ents[0]._.is_family is True

    def test_custom_attribute_error(self):
        """Test that a custom spacy attribute which has not been set
        will throw a ValueError.
        """
        custom_attrs = {
            "FAKE_MODIFIER": {"non_existent_attribute": True},
        }
        with pytest.raises(ValueError):
            ConTextComponent(nlp, add_attrs=custom_attrs)

    def test_custom_attributes_mapping(self):
        custom_attrs = {
            "NEGATED_EXISTENCE": {"is_negated": True},
        }
        try:
            Span.set_extension("is_negated", default=False)
        except:
            pass
        context = ConTextComponent(nlp, add_attrs=custom_attrs)
        assert context.context_attributes_mapping == custom_attrs

    def test_custom_attributes_value1(self):
        custom_attrs = {
            "NEGATED_EXISTENCE": {"is_negated": True},
        }
        try:
            Span.set_extension("is_negated", default=False)
        except:
            pass
        context = ConTextComponent(nlp, add_attrs=custom_attrs)
        context.add([ConTextItem("no evidence of", "NEGATED_EXISTENCE", "FORWARD")])
        doc = nlp("There is no evidence of pneumonia.")
        doc.ents = (doc[-2:-1],)
        context(doc)

        assert doc.ents[0]._.is_negated is True

    def test_custom_attributes_value2(self):
        custom_attrs = {
            "FAMILY": {"is_family": True},
        }
        try:
            Span.set_extension("is_family", default=False)
        except:
            pass
        context = ConTextComponent(nlp, add_attrs=custom_attrs)
        context.add(
            [ConTextItem("no evidence of", "DEFINITE_NEGATED_EXISTENCE", "FORWARD")]
        )
        doc = nlp("There is no evidence of pneumonia.")
        doc.ents = (doc[-2:-1],)
        context(doc)

        assert doc.ents[0]._.is_family is False

    def test_simple_callback(self, capsys):
        context = ConTextComponent(nlp, rules=None)

        def simple_callback(matcher, doc, i, matches):
            match_id, start, end = matches[i]
            span = doc[start:end]
            print("Matched on span:", span)

        context.add(
            [
                ConTextItem(
                    "no evidence of",
                    "NEGATED_EXISTENCE",
                    "FORWARD",
                    on_match=simple_callback,
                )
            ]
        )

        doc = nlp("There is no evidence of pneumonia.")
        context(doc)
        captured = capsys.readouterr()
        assert captured.out == "Matched on span: no evidence of\n"

    def test_global_allowed_types1(self):
        """Check that if the ConTextComponent has allowed_types defined
        and a ConTextItem does not, the ConTextItem will receive the component's
        value.
        """
        context = ConTextComponent(nlp, rules=None, allowed_types={"PROBLEM"})
        item = ConTextItem(
            "no evidence of", "NEGATED_EXISTENCE", "FORWARD", allowed_types=None
        )
        context.add([item])
        assert item.allowed_types == {"PROBLEM"}

    def test_global_allowed_types2(self):
        """Check that if the ConTextComponent does not have allowed_types defined
        and a ConTextItem does, the ConTextItem will not receive the component's
        value.
        """
        context = ConTextComponent(nlp, rules=None, allowed_types=None)
        item = ConTextItem(
            "no evidence of", "NEGATED_EXISTENCE", "FORWARD", allowed_types={"PROBLEM"}
        )
        context.add([item])
        assert item.allowed_types == {"PROBLEM"}

    def test_global_allowed_types2(self):
        """Check that if both the ConTextComponent and a ConTextItem have allowed_types defined,
        the ConTextItem will not receive the component's value.
        """
        context = ConTextComponent(nlp, rules=None, allowed_types={"TREATMENT"})
        item = ConTextItem(
            "no evidence of", "NEGATED_EXISTENCE", "FORWARD", allowed_types={"PROBLEM"}
        )
        context.add([item])
        assert item.allowed_types == {"PROBLEM"}

    def test_context_modifier_termination(self):
        context = ConTextComponent(nlp, rules=None, terminations={"NEGATED_EXISTENCE": ["POSITIVE_EXISTENCE", "UNCERTAIN"]})
        item = ConTextItem(
            "no evidence of", "NEGATED_EXISTENCE", "FORWARD", terminated_by=None
        )
        context.add([item])
        assert item.terminated_by == {"POSITIVE_EXISTENCE", "UNCERTAIN"}

    def test_item_modifier_termination(self):
        context = ConTextComponent(nlp, rules=None,
                                   terminations=None)
        item = ConTextItem(
            "no evidence of", "NEGATED_EXISTENCE", "FORWARD", terminated_by={"POSITIVE_EXISTENCE", "UNCERTAIN"}
        )
        context.add([item])
        assert item.terminated_by == {"POSITIVE_EXISTENCE", "UNCERTAIN"}

    def test_null_modifier_termination(self):
        context = ConTextComponent(nlp, rules=None,
                                   terminations=None)
        item = ConTextItem(
            "no evidence of", "NEGATED_EXISTENCE", "FORWARD", terminated_by=None
        )
        context.add([item])
        assert item.terminated_by == set()

    def test_on_modifies_true(self):
        def on_modifies(target, modifier, span_between):
            return True

        context = ConTextComponent(nlp, rules=None)
        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", on_modifies=on_modifies)
        context.add([item])
        doc = nlp("There is no evidence of pneumonia or chf.")
        doc.ents = (doc[5:6], doc[7:8])
        context(doc)

        for ent in doc.ents:
            assert len(ent._.modifiers) == 1

    def test_on_modifies_false(self):
        def on_modifies(target, modifier, span_between):
            return False

        context = ConTextComponent(nlp, rules=None)
        item = ConTextItem("no evidence of", "NEGATED_EXISTENCE", on_modifies=on_modifies)
        context.add([item])
        doc = nlp("There is no evidence of pneumonia or chf.")
        doc.ents = (doc[5:6], doc[7:8])
        context(doc)

        for ent in doc.ents:
            assert len(ent._.modifiers) == 0

    def test_pseudo_modifier(self):
        item_data = [
            ConTextItem("negative", "NEGATED_EXISTENCE"),
            ConTextItem("negative attitude", "PSEUDO_NEGATED_EXISTENCE", rule="PSEUDO"),
        ]
        context = ConTextComponent(nlp, rules=None)
        context.add(item_data)

        doc = nlp("She has a negative attitude about her treatment.")
        doc.ents = (doc[-2:-1],)
        context(doc)

        assert len(doc.ents[0]._.modifiers) == 0
        assert len(doc._.context_graph.modifiers) == 1
        assert doc._.context_graph.modifiers[0].category == "PSEUDO_NEGATED_EXISTENCE"

    def test_context_window_no_max_scope_fails(self):
        "Test that if use_context_window is True but max_scope is None, the instantiation will fail"
        with pytest.raises(ValueError) as exception_info:
            context = ConTextComponent(nlp, max_scope=None, use_context_window=True)
        exception_info.match(
            "If 'use_context_window' is True, 'max_scope' must be an integer greater 1, not None"
        )



import pytest

import medspacy
import spacy


class TestMedSpaCy:
    def test_default_build_pipe_names(self):
        model, enable, disable = medspacy.util._build_pipe_names("default", enable=None, disable=None)
        assert model == "en_core_web_sm"
        assert enable == {"tokenizer", "sentencizer", "target_matcher", "context"}
        assert disable == {"parser", "tagger", "ner"}

    def test_default_load(self):
        nlp = medspacy.load()
        expected_pipe_names = {
            "sentencizer",
            "context",
            "target_matcher",
        }
        assert set(nlp.pipe_names) == expected_pipe_names

    def test_load_enable(self):
        nlp = medspacy.load(enable={"target_matcher", "sectionizer"})
        assert len(nlp.pipeline) == 2
        assert set(nlp.pipe_names) == {"target_matcher", "sectionizer"}

    def test_nlp(self):
        nlp = medspacy.load()
        assert nlp("This is a sentence. So is this.")

    def test_load_disable(self):
        nlp = medspacy.load(disable=["context"])
        expected_pipe_names = {
            "sentencizer",
            "target_matcher",
        }
        assert set(nlp.pipe_names) == expected_pipe_names

    def test_load_all_components(self):
        full_pipe_names = [
            "sentencizer",
            "target_matcher",
            "context",
            "sectionizer",
            "postprocessor"
        ]

        nlp = medspacy.load(enable="all")
        assert nlp.pipe_names == full_pipe_names
        assert isinstance(nlp.tokenizer, medspacy.preprocess.Preprocessor)

    # def test_load_sci(self):
    #     # pip install https://s3-us-west-2.amazonaws.com/ai2-s2-scispacy/releases/v0.3.0/en_core_sci_sm-0.3.0.tar.gz
    #     assert medspacy.load("en_core_sci_sm")

    def test_load_rules(self):
        nlp = medspacy.load(load_rules=True)
        context = nlp.get_pipe("context")
        assert context.item_data

    def test_not_load_rules(self):
        nlp = medspacy.load(load_rules=False)
        context = nlp.get_pipe("context")
        assert not context.item_data

    def test_medspacy_tokenizer(self):
        default_tokenizer = spacy.blank("en").tokenizer
        custom_tokenizer = medspacy.load(enable=['tokenizer']).tokenizer

        text = r'Pt c\o n;v;d h\o chf+cp n/v/d'

        default_doc = default_tokenizer(text)
        medspacy_doc = custom_tokenizer(text)

        assert [token.text for token in default_doc] != [token.text for token in medspacy_doc]

        # Check that some expected token boundries are generated
        joined_tokens = " ".join([token.text for token in medspacy_doc])
        assert "c \\ o" in joined_tokens
        assert "n / v / d" in joined_tokens
        assert "chf + cp" in joined_tokens

    def test_disable_medspacy_tokenizer(self):
        default_tokenizer = spacy.blank("en").tokenizer
        custom_tokenizer = medspacy.load(disable=['tokenizer']).tokenizer

        text = r'Pt c\o n;v;d h\o chf+cp n/v/d'

        default_doc = default_tokenizer(text)
        medspacy_doc = custom_tokenizer(text)

        assert [token.text for token in default_doc] == [token.text for token in medspacy_doc]

    def test_medspacy_tokenizer_uppercase(self):
        custom_tokenizer = medspacy.load(enable=['tokenizer']).tokenizer

        # Issue 13: Ensure that uppercase tokens are not tokenized as each character
        # https://github.com/medspacy/medspacy/issues/13
        text = r'DO NOT BREAK ME UP'

        medspacy_doc = custom_tokenizer(text)

        tokens = [token.text for token in medspacy_doc]

        assert len(tokens) == 5

        # Check that some expected token boundries are generated
        joined_tokens = " ".join(tokens)
        assert "DO NOT BREAK ME UP" in joined_tokens
        assert "B R E A K" not in joined_tokens

    def test_medspacy_tokenizer_numerics(self):
        custom_tokenizer = medspacy.load(enable=['tokenizer']).tokenizer

        text = r'1.5 mg'

        medspacy_doc = custom_tokenizer(text)

        tokens = [token.text for token in medspacy_doc]

        assert len(tokens) == 2

        # Check that some expected token boundries are generated
        joined_tokens = " ".join(tokens)
        assert "1.5" in joined_tokens
        assert "1 . 5" not in joined_tokens




"""This module will contain helper functions and classes for common clinical processing tasks
which will be done in conjunction with cycontext.
"""


def is_modified_by(span, modifier_label):
    for modifier in span._.modifiers:
        if modifier.category.upper() == modifier_label.upper():
            return True
    return False



from .context_component import ConTextComponent, DEFAULT_RULES_FILEPATH, DEFAULT_ATTRS
from .context_item import ConTextItem
from ._version import __version__

__all__ = ["ConTextComponent", "ConTextItem", "DEFAULT_RULES_FILEPATH"]



NEWLINE_PATTERN = r"[\n\r]+[\s]*$"


def get_section_titles(doc):
    return [title for (title, _, _, _) in doc._.sections]


def get_section_headers(doc):
    return [header for (_, header, _, _) in doc._.sections]


def get_section_parents(doc):
    return [parent for (_, _, parent, _,) in doc._.sections]


def get_section_spans(doc):
    return [span for (_, _, _, span) in doc._.sections]


def is_start_line(idx, doc, pattern):
    # If it's the start of the doc, return True
    if idx == 0:
        return True
    # Otherwise, check if the preceding token ends with newlines
    preceding_text = doc[idx - 1].text_with_ws
    return pattern.search(preceding_text) is not None


def is_end_line(idx, doc, pattern):
    # If it's the end of the doc, return True
    if idx == len(doc) - 1:
        return True

    # Check if either the token has trailing newlines,
    # or if the next token is a newline
    text = doc[idx].text_with_ws
    if pattern.search(text) is not None:
        return True
    following_text = doc[idx + 1].text_with_ws
    return pattern.search(following_text) is not None



class TagObject:
    """Represents a concept found by ConText in a document.
    Is the result of ConTextItem matching a span of text in a Doc.
    """

    def __init__(
        self, context_item, start, end, doc, _use_context_window=False
    ):
        """Create a new TagObject from a document span.

        context_item (int): The ConTextItem object which defines the modifier.
        start (int): The start token index.
        end (int): The end token index (non-inclusive).
        doc (Doc): The spaCy Doc which contains this span.
        """
        self.context_item = context_item
        self.start = start
        self.end = end
        self.doc = doc

        self._targets = []
        self._num_targets = 0

        self._use_context_window = _use_context_window
        self._scope_start = None
        self._scope_end = None

        self.set_scope()

    @property
    def span(self):
        """The spaCy Span object, which is a view of self.doc, covered by this match."""
        return self.doc[self.start : self.end]

    @property
    def rule(self):
        """Returns the associated rule."""
        return self.context_item.rule

    @property
    def category(self):
        """Returns the associated category."""
        return self.context_item.category

    @property
    def scope(self):
        """Returns the associated scope."""
        return self.doc[self._scope_start : self._scope_end]

    @property
    def allowed_types(self):
        """Returns the associated allowed types."""
        return self.context_item.allowed_types

    @property
    def excluded_types(self):
        """Returns the associated excluded types."""
        return self.context_item.excluded_types

    @property
    def num_targets(self):
        """Returns the associated number of targets."""
        return self._num_targets

    @property
    def max_targets(self):
        """Returns the associated maximum number of targets."""
        return self.context_item.max_targets

    @property
    def max_scope(self):
        """Returns the associated maximum scope."""
        return self.context_item.max_scope

    def set_scope(self):
        """Applies the rule of the ConTextItem which generated
        this TagObject to define a scope.
        If self.max_scope is None, then the default scope is the sentence which it occurs in
        in whichever direction defined by self.rule.
        For example, if the rule is "forward", the scope will be [self.end: sentence.end].
        If the rule is "backward", it will be [self.start: sentence.start].

        If self.max_scope is not None and the length of the default scope is longer than self.max_scope,
        it will be reduced to self.max_scope.


        """
        # If ConText is set to use defined windows, do that instead of sentence splitting
        if self._use_context_window:
            # Up to the beginning of the doc
            full_scope_start = max(
                (0, self.start - self.context_item.max_scope)
            )
            # Up to the end of the doc
            full_scope_end = min(
                (len(self.span.doc), self.end + self.context_item.max_scope)
            )
            full_scope_span = self.span.doc[full_scope_start:full_scope_end]
        # Otherwise, use the sentence
        else:
            full_scope_span = self.doc[self.start].sent
            if full_scope_span is None:
                raise ValueError(
                    "ConText failed because sentence boundaries have not been set and 'use_context_window' is set to False. "
                    "Add an upstream component such as the dependency parser, Sentencizer, or PyRuSH to detect sentence "
                    "boundaries or initialize ConTextComponent with 'use_context_window=True.'"
                )

        if self.rule.lower() == "forward":
            self._scope_start, self._scope_end = self.end, full_scope_span.end
            if (
                self.max_scope is not None
                and (self._scope_end - self._scope_start) > self.max_scope
            ):
                self._scope_end = self.end + self.max_scope

        elif self.rule.lower() == "backward":
            self._scope_start, self._scope_end = (
                full_scope_span.start,
                self.start,
            )
            if (
                self.max_scope is not None
                and (self._scope_end - self._scope_start) > self.max_scope
            ):
                self._scope_start = self.start - self.max_scope
        else:  # bidirectional
            self._scope_start, self._scope_end = (
                full_scope_span.start,
                full_scope_span.end,
            )

            # Set the max scope on either side
            # Backwards
            if (
                self.max_scope is not None
                and (self.start - self._scope_start) > self.max_scope
            ):
                self._scope_start = self.start - self.max_scope
            # Forwards
            if (
                self.max_scope is not None
                and (self._scope_end - self.end) > self.max_scope
            ):
                self._scope_end = self.end + self.max_scope

    def update_scope(self, span):
        """Change the scope of self to be the given spaCy span.

        span (Span): a spaCy Span which contains the scope
        which a modifier should cover.
        """
        self._scope_start, self._scope_end = span.start, span.end

    def limit_scope(self, other):
        """If self and obj have the same category
        or if obj has a directionality of 'terminate',
        use the span of obj to update the scope of self.
        Limiting the scope of two modifiers of the same category
        reduces the number of modifiers. For example, in
        'no evidence of CHF, no pneumonia', 'pneumonia' will only
        be modified by 'no', not 'no evidence of'.
        'terminate' modifiers limit the scope of a modifier
        like 'no evidence of' in 'no evidence of CHF, **but** there is pneumonia'

        other (TagObject)
        Returns True if obj modfified the scope of self
        """
        if self.span.sent != other.span.sent:
            return False
        if self.rule.upper() == "TERMINATE":
            return False
        # Check if the other modifier is a type which can modify self
        # or if they are the same category. If not, don't reduce scope.
        if (
            (other.rule.upper() != "TERMINATE")
            and (other.category.upper() not in self.context_item.terminated_by)
            and (other.category.upper() != self.category.upper())
        ):
            return False

        # If two modifiers have the same category but modify different target types,
        # don't limit scope.
        if self.category == other.category and (
            (self.allowed_types != other.allowed_types)
            or (self.excluded_types != other.excluded_types)
        ):
            return False

        orig_scope = self.scope
        if self.rule.lower() in ("forward", "bidirectional"):
            if other > self:
                self._scope_end = min(self._scope_end, other.start)
        if self.rule.lower() in ("backward", "bidirectional"):
            if other < self:
                self._scope_start = max(self._scope_start, other.end)
        return orig_scope != self.scope

    def modifies(self, target):
        """Returns True if the target is within the modifier scope
        and self is allowed to modify target.

        target (Span): a spaCy span representing a target concept.
        """
        # If the target and modifier overlap, meaning at least one token
        # one extracted as both a target and modifier, return False
        # to avoid self-modifying concepts

        if self.overlaps_target(target):
            return False
        if self.rule in ("TERMINATE", "PSEUDO"):
            return False
        if not self.allows(target.label_.upper()):
            return False

        if target[0] in self.scope or target[-1] in self.scope:
            if not self.on_modifies(target):
                return False
            else:
                return True
        return False

    def allows(self, target_label):
        """Returns True if a modifier is able to modify a target type.
        A modifier may not be allowed if either self.allowed_types is not None and
        target_label is not in it, or if self.excluded_types is not None and
        target_label is in it.
        """
        if self.allowed_types is not None:
            if target_label not in self.allowed_types:
                return False
            return True
        if self.excluded_types is not None:
            if target_label not in self.excluded_types:
                return True
            return False
        return True

    def on_modifies(self, target):
        """If the ConTextItem used to define a TagObject has an on_modifies callback function,
        evaluate and return either True or False.
        If on_modifies is None, return True.
        """
        if self.context_item.on_modifies is None:
            return True
        # Find the span in between the target and modifier
        start = min(target.end, self.span.end)
        end = max(target.start, self.span.start)
        span_between = target.doc[start:end]
        rslt = self.context_item.on_modifies(target, self.span, span_between)
        if rslt not in (True, False):
            raise ValueError(
                "The on_modifies function must return either True or False indicating "
                "whether a modify modifies a target. Actual value: {0}".format(
                    rslt
                )
            )
        return rslt

    def modify(self, target):
        """Add target to the list of self._targets and increment self._num_targets."""
        self._targets.append(target)
        self._num_targets += 1

    def reduce_targets(self):
        """If self.max_targets is not None, reduce the targets which are modified
        so that only the n closest targets are left. Distance is measured as
        the distance to either the start or end of a target (whichever is closer).
        """
        if self.max_targets is None or self.num_targets <= self.max_targets:
            return

        target_dists = []
        for target in self._targets:
            dist = min(
                abs(self.start - target.end), abs(target.start - self.end)
            )
            target_dists.append((target, dist))
        srtd_targets, _ = zip(*sorted(target_dists, key=lambda x: x[1]))
        self._targets = srtd_targets[: self.max_targets]
        self._num_targets = len(self._targets)

    def overlaps(self, other):
        """ Returns whether the object overlaps with another span

        other (): the other object to check for overlaps

        RETURNS: true if there is overlap, false otherwise.
        """
        return (
            self.span[0] in other.span
            or self.span[-1] in other.span
            or other.span[0] in self.span
            or other.span[-1] in self.span
        )

    def overlaps_target(self, target):
        """Returns True if self overlaps with a spaCy span."""
        return (
            self.span[0] in target
            or self.span[-1] in target
            or target[0] in self.span
            or target[-1] in self.span
        )

    def __gt__(self, other):
        return self.span > other.span

    def __ge__(self, other):
        return self.span >= other.span

    def __lt__(self, other):
        return self.span < other.span

    def __le__(self, other):
        return self.span <= other.span

    def __len__(self):
        return len(self.span)

    def __repr__(self):
        return f"<TagObject> [{self.span}, {self.category}]"
